import warnings
import pandas as pd
import logging
from ._version import __version__

warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)

# level = logging.INFO
# spoint_logger = logging.getLogger("spoint")
# spoint_logger.propagate = False
# spoint_logger.setLevel(level)
# if len(spoint_logger.handlers) == 0:
#     console = Console(force_terminal=True)
#     if console.is_jupyter is True:
#         console.is_jupyter = False
#     ch = RichHandler(
#         level=level, show_path=False, console=console, show_time=False
#     )
#     formatter = logging.Formatter("%(message)s")
#     ch.setFormatter(formatter)
#     spoint_logger.addHandler(ch)



import numpy as np
import pandas as pd
import torch
import gpytorch
import math
import os
import scanpy as sc
from .plot import plot_3d

class ExactGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood,lengthscale_prior=None,outputscale_prior=None):
        super(ExactGPModel, self).__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = gpytorch.kernels.ScaleKernel(
            base_kernel=gpytorch.kernels.RBFKernel(
                lengthscale_prior=lengthscale_prior
            ),
            outputscale_prior=outputscale_prior
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

class GPRmodel():
    def __init__(self,expr,loc,loc_resample,used_genes,use_gpu=False,output_dir=None,**kwargs):
        
        if 'subset' not in kwargs.keys():
            self.subset = 10000
        else:
            self.subset = kwargs['subset']
            
        if 'lengthscale_prior' not in kwargs.keys():
            self.lengthscale_prior = None
        else:
            self.lengthscale_prior = kwargs['lengthscale_prior']
            
        if 'outputscale_prior' not in kwargs.keys():
            self.outputscale_prior = None
        else:
            self.outputscale_prior = kwargs['outputscale_prior']
            
        if 'noise_prior' not in kwargs.keys():
            self.noise_prior = None
        else:
            self.noise_prior = kwargs['noise_prior']
        
        if output_dir is not None:
            self.output_dir = output_dir
        else:
            self.output_dir = './gpr_models'
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
            
        self.model = None
        self.flat_model = None
        self.loc_resample = torch.tensor(loc_resample,dtype=torch.float32)
        self.train_x, self.subset_y = self.prepare_gpr_data(loc,expr.values,self.subset)
        self.train_y = None
        self.used_genes = used_genes
        self.g_ind = [list(expr.columns).index(g) for g in self.used_genes]
        self.gene_ind_map = pd.DataFrame(self.g_ind,index=self.used_genes,columns=['g_ind'])
        self.log_bf = pd.DataFrame(index=self.used_genes,columns=['log_bf'])
        self.use_gpu = use_gpu
        if self.use_gpu:
            print('Using GPU accelerate')
            self.train_x = self.train_x.cuda()
            self.loc_resample = self.loc_resample.cuda()
    
    @staticmethod   
    def prepare_gpr_data(X,y,subset):
        subset_ind = np.random.permutation(X.shape[0])[:subset]
        subset_x = X[subset_ind]
        subset_y = y[subset_ind]
        subset_x = torch.tensor(subset_x,dtype=torch.float32)
        subset_y = torch.tensor(subset_y,dtype=torch.float32)
        return subset_x, subset_y
    
    def prepare_gpr_model(self, lengthscale_prior=None,outputscale_prior=None,noise_prior=None,bayesian_alter=False):

        if bayesian_alter:
            if self.flat_model is None:
                likelihood = gpytorch.likelihoods.GaussianLikelihood(
                    noise_prior=noise_prior
                )
                self.flat_model = ExactGPModel(self.train_x, self.train_y, likelihood, lengthscale_prior=lengthscale_prior,outputscale_prior=outputscale_prior)
            # self.init_model(self.flat_model,lengthscale=torch.tensor(99999999))
            self.init_model(self.flat_model,lengthscale=torch.inf)
        else:
            if self.model is None:
                likelihood = gpytorch.likelihoods.GaussianLikelihood(
                    noise_prior=noise_prior
                )
                self.model = ExactGPModel(self.train_x, self.train_y, likelihood, lengthscale_prior=lengthscale_prior,outputscale_prior=outputscale_prior)
            self.init_model(self.model)
            
    
    def init_model(self,model,noise=None,lengthscale=None,outputscale=None,constant=None):
        if noise is None:
            noise = self.train_y.std()/2
        if lengthscale is None:
            lengthscale = self.train_y.std()
        if outputscale is None:
            outputscale = torch.tensor(4)
        if constant is None:
            constant = torch.tensor(self.train_y.mean())
            
        hypers = {
            'likelihood.noise_covar.noise': noise,
            'covar_module.base_kernel.lengthscale': lengthscale,
            'covar_module.outputscale': outputscale,
            'mean_module.constant': constant
        }

        model.initialize(**hypers)
        model.train_targets = self.train_y
        model.zero_grad()
        # print('outputscale: %.3f   lengthscale: %.3f   noise: %.3f' % (
        #     model.covar_module.outputscale.item(),
        #     model.covar_module.base_kernel.lengthscale.item(),
        #     model.likelihood.noise.item()
        # ))
    
    def train_single_model(self,model,lr=1,training_iter=500,save=False,save_path=None,optimize_method='Adam'):
        # Find optimal model hyperparameters
        model.train()
        model.likelihood.train()
        
        # "Loss" for GPs - the marginal log likelihood
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(model.likelihood, model)
        best_loss=np.inf
        
        if optimize_method=='Adam':
            optimizer = torch.optim.Adam(model.parameters(),lr=lr)
            for i in range(training_iter):
                optimizer.zero_grad()
                output = model(self.train_x)
                loss = -mll(output, self.train_y)
                loss.backward()
                optimizer.step()
                # print('Iter %d/%d - Loss: %.3f   lengthscale: %.3f   noise: %.3f' % (
                #     i + 1, training_iter, loss.item(),
                #     model.covar_module.base_kernel.lengthscale.item(),
                #     model.likelihood.noise.item()
                # ))
                if loss.item() < best_loss:
                    best_model_state = model.state_dict()
                    best_loss = loss.item()
        # Bugs need to be solved
        elif optimize_method=='LBGFS':
            optimizer = torch.optim.LBFGS(model.parameters(),line_search_fn='strong_wolfe', lr=lr)
            def closure():
                optimizer.zero_grad()
                output = model(self.train_x)
                loss = -mll(output, self.train_y)
                loss.backward()
                return loss
            for i in range(training_iter):
                loss = optimizer.step(closure)
                # print('Iter %d/%d - Loss: %.3f   lengthscale: %.3f   noise: %.3f' % (
                #     i + 1, training_iter, loss.item(),
                #     model.covar_module.base_kernel.lengthscale.item(),
                #     model.likelihood.noise.item()
                # ))
                if loss.item() < best_loss:
                    best_model_state = model.state_dict()
                    best_loss = loss.item()
        else:
            raise ValueError('Invalid optimize method: ', optimize_method)

        # Get into evaluation (predictive posterior) mode
        model.eval()
        model.likelihood.eval()
        # for param_name, param in model.named_parameters():
        #     print(f'Parameter name: {param_name:42} value = {param.item()}')
        if save:
            torch.save(best_model_state, save_path)
        print('Best model loss:', best_loss)
        return best_loss
    
    def optim_lengthscale(self,model,lr=1,l_range=torch.arange(1,12,1),optimize_method='Adam'):

        model.train()
        model.likelihood.train()
        loss_list = []
        print('Optimize lenthscale...')
        for lenthscale_alpha in l_range:
            self.init_model(model,lengthscale=lenthscale_alpha*self.train_y.std())
            mll = gpytorch.mlls.ExactMarginalLogLikelihood(model.likelihood, model)
            
            if optimize_method=='Adam':
                optimizer = torch.optim.Adam(model.parameters(),lr=lr)
                optimizer.zero_grad()
                output = model(self.train_x)
                loss = -mll(output, self.train_y)
                loss.backward()
                optimizer.step()
            elif optimize_method=='LBGFS':
                optimizer = torch.optim.LBFGS(model.parameters(),line_search_fn='strong_wolfe', lr=lr)
                def closure():
                    optimizer.zero_grad()
                    output = model(self.train_x)
                    loss = -mll(output, self.train_y)
                    loss.backward()
                    return loss
                loss = optimizer.step(closure)
            else:
                raise ValueError('Invalid optimize method: ', optimize_method)
            # print(' - Lengthscale weight: %.3f   Loss: %.3f' % (
            #     lenthscale_alpha,
            #     loss.item(),
            # ))
            loss_list.append(loss.item())
        best_l_alpha = l_range[np.argmin(loss_list)]
        print('Initialize lenthscale alpha as %.3f' % best_l_alpha)
        self.init_model(model,lengthscale=best_l_alpha*self.train_y.std())
        model.eval()
        model.likelihood.eval()
    
    def train(self,lr=1,training_iter=500,save_model=True,save_pred=False,cal_bf=False,optim_l=False,optimize_method='Adam'):
        for g,g_i in zip(self.used_genes,self.g_ind):
            print(f'Modeling {g}')
            self.train_y = self.subset_y[:,g_i]
            self.prepare_gpr_model(lengthscale_prior=self.lengthscale_prior,outputscale_prior=self.outputscale_prior,noise_prior=self.noise_prior) 
            if self.use_gpu:
                self.train_y = self.train_y.cuda()
                self.model = self.model.cuda()
                self.model.likelihood = self.model.likelihood.cuda()
                
            if optim_l:
                self.optim_lengthscale(self.model,lr=lr,optimize_method=optimize_method)
            
            mll = self.train_single_model(
                self.model,
                lr=lr,
                optimize_method=optimize_method,
                training_iter=training_iter,
                save=save_model,
                save_path=os.path.join(self.output_dir,f'{g}_iter{training_iter}_lr{lr}_model_state.pth'))
            
            if save_pred:
                resampled_pred = self.predict_resampled_spot()
                np.save(os.path.join(self.output_dir,f'{g}_iter{training_iter}_lr{lr}_resampled_pred.npy'),resampled_pred)
            
            if cal_bf:
                self.prepare_gpr_model(lengthscale_prior=self.lengthscale_prior,outputscale_prior=self.outputscale_prior,noise_prior=self.noise_prior,bayesian_alter=True) 
                if self.use_gpu:
                    self.flat_model = self.flat_model.cuda()
                    self.flat_model.likelihood = self.flat_model.likelihood.cuda()
                flat_mll = self.train_single_model(self.flat_model,lr=lr,training_iter=training_iter,save=save_model,save_path=os.path.join(self.output_dir,f'{g}_iter{training_iter}_lr{lr}_flat_model_state.pth'))
                log_bf = -mll+flat_mll
                self.log_bf.loc[g,'log_bf'] = log_bf
                print(log_bf)

    def load_gene_model(self,gene,training_iter,lr):
        g_i = self.gene_ind_map.loc[gene,'g_ind']
        self.train_y = self.subset_y[:,g_i]
        if self.use_gpu:
            self.train_y = self.train_y.cuda()
        if self.model is None:
            self.prepare_gpr_model()
        state_dict = torch.load(os.path.join(self.output_dir,f'{gene}_iter{training_iter}_lr{lr}_model_state.pth'))
        self.model.load_state_dict(state_dict)
        self.model.train_targets = self.train_y

    def eval_model(self):
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            observed_pred = self.model.likelihood(self.model(self.train_x))
        return observed_pred
    
    def predict_resampled_spot(self, gene=None, data=None, training_iter=None, lr=None, save_pred=False, save_pred_path=None, n_in_batch=30000):
        if gene is not None:
            self.load_gene_model(gene, training_iter, lr)
        obs_pred_list = []
        for i in range(int(np.ceil(data.shape[0]/n_in_batch))):
            low_bound = i*n_in_batch
            high_bound = min((i+1)*n_in_batch,data.shape[0])
            data_batch = data[low_bound:high_bound,:]
            with torch.no_grad(), gpytorch.settings.fast_pred_var():
                observed_pred_batch = self.model.likelihood(self.model(torch.tensor(data_batch)))
            obs_pred_list.append(observed_pred_batch.mean.cpu().numpy())
        obs_pred_concated = np.concatenate(obs_pred_list)
        if save_pred:
            if save_pred_path is None:
                save_pred_path = os.path.join(self.output_dir,f'{gene}_iter{training_iter}_lr{lr}_resampled_pred.npy')
            np.save(save_pred_path,obs_pred_concated)
        return obs_pred_concated
    
    def plot_gpr_expr(
        self,
        gene,
        training_iter,
        lr,
        data=None,
        pred_path=None,
        save=False,
        save_path=None,
        save_pred=False,
        save_pred_path=None,
        save_dpi=150,
        return_expr=True,
        *args,**kwargs
    ):
        if data is None:
            data = self.loc_resample
        else:
            data = torch.tensor(data,dtype=torch.float32)
            if self.use_gpu:
                data = data.cuda()
        # if pred_path is None:
        #     pred_path = os.path.join(self.output_dir,f'{gene}_iter{training_iter}_lr{lr}_resampled_pred.npy')
        if save and save_path is None:
            save_path = os.path.join(self.output_dir,f'{gene}_iter{training_iter}_lr{lr}_resampled_pred.png')
        if (pred_path is not None) and os.path.exists(pred_path):
            resampled_pred = np.load(os.path.join(self.output_dir,f'{gene}_iter{training_iter}_lr{lr}_resampled_pred.npy'))
        else:
            resampled_pred = self.predict_resampled_spot(gene,data,training_iter,lr,save_pred,save_pred_path)
        plot_3d(data.cpu().numpy(), val=resampled_pred,save_path=save_path,save_dpi=save_dpi,*args,**kwargs)
        if return_expr:
            return resampled_pred



import os 
import numpy as np
import pandas as pd
from multiprocessing import Pool
from scipy.optimize import minimize, differential_evolution
from sklearn.metrics import pairwise_distances
from sklearn.neighbors import BallTree
from copy import deepcopy
import time
from functools import partial

def rotate_loc(x,y,x0,y0,angle):
    angle = angle/(180/np.pi)
    x1=np.cos(angle)*(x-x0)-np.sin(angle)*(y-y0) + x0
    y1=np.cos(angle)*(y-y0)+np.sin(angle)*(x-x0) + y0
    return x1,y1

# Deprecated
def neighbor_simi(source_loc,target_loc, source_cluster, target_cluster, target_p_dis,k,knn_exclude_cutoff):
    dis_knn_exclude_cutoff = np.median(np.partition(target_p_dis,kth=k+knn_exclude_cutoff,axis=1)[:,k+knn_exclude_cutoff])
    p_dis = pairwise_distances(source_loc,target_loc)
    top5_neighbors = np.argpartition(p_dis,kth=k,axis=1)[:,:k]
    n_simi = []
    for i, sc in enumerate(source_cluster):
        tmp_neighbors = list(top5_neighbors[i])
        tmp_neighbors_reversed = tmp_neighbors[::-1]
        for j in tmp_neighbors_reversed:
            if p_dis[i,j] > dis_knn_exclude_cutoff:
                tmp_neighbors.remove(j)
            else:
                break
        if len(tmp_neighbors) != 0: 
            n_simi.append((target_cluster[tmp_neighbors] == sc).sum()/len(tmp_neighbors))
    return np.mean(n_simi) - (len(n_simi)/i-1)**2

def neighbor_simi_fast(source_loc,target_loc, source_cluster, target_cluster,k,knn_exclude_cutoff,p=2,a=1):
    target_tree = BallTree(target_loc, leaf_size=15, metric='minkowski')
    t_dis, _ = target_tree.query(target_loc, k=k+knn_exclude_cutoff)
    dis_knn_exclude_cutoff = np.median(t_dis[:,k+knn_exclude_cutoff-1])
    s_dis, nearest_neighbors = target_tree.query(source_loc, k=k)
    nearest_neighbors[s_dis > dis_knn_exclude_cutoff] = -1
    kept_ind = (nearest_neighbors != -1).sum(1)>0
    if kept_ind.sum() > 0:
        nearest_neighbors = nearest_neighbors[kept_ind]
        neighbors_cluster = deepcopy(nearest_neighbors)
        neighbors_cluster[nearest_neighbors!=-1] = target_cluster[nearest_neighbors[nearest_neighbors!=-1]]
        simi = (((neighbors_cluster - source_cluster[kept_ind].reshape(-1,1)) == 0).sum(1)/(neighbors_cluster!=-1).sum(1)).mean()
    else:
        simi = 0
    overlap = kept_ind.sum()/len(source_cluster)
    neighbors_simi = simi - a*((1-overlap)**p)
    return neighbors_simi

def neighbor_simi_new(source_loc,target_loc, source_cluster, target_cluster,k,knn_exclude_cutoff,p=2,a=1):
    target_tree = BallTree(target_loc, leaf_size=15, metric='minkowski')
    source_tree = BallTree(source_loc, leaf_size=15, metric='minkowski')
    t_dis, _ = target_tree.query(target_loc, k=k+knn_exclude_cutoff)
    dis_knn_exclude_cutoff = np.median(t_dis[:,k+knn_exclude_cutoff-1])
    s_dis, target_nearest_neighbors = target_tree.query(source_loc, k=k)
    _, source_nearest_neighbors = source_tree.query(source_loc, k=k)
    target_nearest_neighbors[s_dis > dis_knn_exclude_cutoff] = -1
    kept_ind = (target_nearest_neighbors != -1).sum(1)>0
    if kept_ind.sum() > 0:
        target_nearest_neighbors = target_nearest_neighbors[kept_ind]
        target_neighbors_cluster = deepcopy(target_nearest_neighbors)
        target_neighbors_cluster[target_nearest_neighbors!=-1] = target_cluster[target_nearest_neighbors[target_nearest_neighbors!=-1]]
        source_neighbors_cluster = source_cluster[source_nearest_neighbors][kept_ind]
        cluster_num = max(np.unique(np.concatenate([target_cluster,source_cluster])))+1
        cluster_onehot = np.concatenate([np.eye(cluster_num),np.zeros((1,cluster_num))])
        # simi = (((neighbors_cluster - source_cluster[kept_ind].reshape(-1,1)) == 0).sum(1)/(neighbors_cluster!=-1).sum(1)).mean()
        target_cluster_count = cluster_onehot[target_neighbors_cluster].sum(1)
        source_cluster_count = cluster_onehot[source_neighbors_cluster].sum(1)
        target_cluster_prop = target_cluster_count/target_cluster_count.sum(1,keepdims=True)
        source_cluster_prop = source_cluster_count/source_cluster_count.sum(1,keepdims=True)
        simi = ((source_cluster_prop - target_cluster_prop)**2).sum(1).mean()/2
    else:
        simi = 1
    overlap = kept_ind.sum()/len(source_cluster)
    neighbors_simi = simi + a*((1-overlap)**p)
    # print(simi,overlap)
    return -neighbors_simi

def score(warp_param, target_loc, source_loc, target_cluster, source_cluster,k,knn_exclude_cutoff,p,a):
    new_source_loc = []
    x0 = 0
    y0 = 0
    for x,y in source_loc:
        new_source_loc.append(rotate_loc(x,y,x0,y0,warp_param[0]))
    new_source_loc = np.array(new_source_loc)
    new_source_loc[:,0] += warp_param[1]  #x方向平移
    new_source_loc[:,1] += warp_param[2]  #y方向平移
    return -neighbor_simi_fast(new_source_loc, target_loc, source_cluster, target_cluster, k, knn_exclude_cutoff,p,a)

def optimize(target_loc, source_loc, target_cluster, source_cluster, n_neighbors, knn_exclude_cutoff ,p,a, bound_alpha,n_threads,*args,**kwargs): 

    source_loc_flip = deepcopy(source_loc)
    source_loc_flip[:,0] = -source_loc_flip[:,0]
    func1 = partial(score, target_loc=target_loc, source_loc=source_loc, target_cluster=target_cluster, source_cluster=source_cluster,k=n_neighbors,knn_exclude_cutoff=knn_exclude_cutoff,p=p,a=a)
    opm1 = differential_evolution(func1, bounds=((0, 360), (target_loc.min(0)[0]*bound_alpha, target_loc.max(0)[0]*bound_alpha), (target_loc.min(0)[1]*bound_alpha, target_loc.max(0)[1]*bound_alpha)),workers=n_threads,*args,**kwargs)
    score1 = -opm1.fun
    result1 = opm1.x
    
    func2 = partial(score, target_loc=target_loc, source_loc=source_loc_flip, target_cluster=target_cluster, source_cluster=source_cluster,k=n_neighbors,knn_exclude_cutoff=knn_exclude_cutoff,p=p,a=a)
    opm2 = differential_evolution(func2,bounds=((0, 360), (target_loc.min(0)[0]*bound_alpha, target_loc.max(0)[0]*bound_alpha), (target_loc.min(0)[1]*bound_alpha, target_loc.max(0)[1]*bound_alpha)),workers=n_threads,*args,**kwargs)
    score2 = -opm2.fun
    result2 = opm2.x

    print(f'Optimal score: flip = {score1}, not flip = {score2}')

    if score1 > score2:
        return 0, result1, score1, 1, result2, score2
    else:
        return 1, result2, score2, 0, result1, score1

def align_pairwise(param, n_neighbors, knn_exclude_cutoff,p,a,bound_alpha,n_threads,*args,**kwargs):
    i, target_loc,source_loc,target_cluster,source_cluster = param
    flip, result, score, alter_flip, alter_result,alter_score  = optimize(target_loc, source_loc, target_cluster, source_cluster, n_neighbors, knn_exclude_cutoff, p,a,bound_alpha,n_threads,*args,**kwargs)
    return [i,flip,result[0],result[1],result[2],score,alter_flip,alter_result[0],alter_result[1],alter_result[2],alter_score]

def align(
    ad_list,
    cluster_key='spatial_domain',
    output_path=None,
    raw_loc_key='spatial',
    aligned_loc_key='spatial_aligned',
    n_neighbors=5,
    knn_exclude_cutoff=1,
    p=2,
    a=1,
    bound_alpha=1,
    write_loc_path=None,
    n_threads=1,
    seed=42,
    subset_prop=None,
    *args,
    **kwargs
):
    """Pairwise alignment.
    
    Pairwise align the slices in ad_list. The aligned coordinates are saved in ``.obsm[aligned_loc_key]`` in each slices of ``ad_list``.

    Args:
        ad_list: A list containing all slice data in AnnData object.
        cluster_key: A string representing one column of ``obs`` in AnnData object, containing the spatial domain information used for alignment.
        output_path: A string representing the path directory where the alignment parameters are saved. If ``None``, it will be 'Scube_outputs'.
        raw_loc_key: A string representing one key of ``obsm`` in AnnData object of each slices in ``ad_list``, containing the raw coordinates.
        aligned_loc_key: A string written to a key of ``obsm`` in AnnData object of each slices in ``ad_list``, containing the aligned coordinates.
        n_neighbors: A number of neighbors in target slices considered by each spot/cell in source slices.
        knn_exclude_cutoff: A number used to filter the neighbors in KNN. The neighbor will be exclude when the distance of neighbors larger than the median of neareast ``n_neighbors + knn_exclude_cutoff`` neighbors distance in all spot/cell in target slice.
        p: Degree of the penalty function.
        a: Coefficient of the penalty function.
        bound_alpha: For the optimized boundary, the multiplier based on the maximum and minimum values of the slice coordinates.
        write_loc_path: A string representing the path directory where the aligned coordinates of all slices are saved. If ``None``, it won't be saved.
        n_threads: The number of parallel threads for the optimization algorithm.
        seed: Seed for the optimization algorithm.
        subset_prop: The downsampling ratio for cells in each slice.
    
    Returns:
        ``None``
    """
    if subset_prop is not None:
        for i in range(len(ad_list)):
            ad_list[i] = ad_list[i][np.random.permutation(ad_list[i].obs_names)[:int(ad_list[i].shape[0]*subset_prop)]].copy()
    if output_path is None:
        output_path = 'Scube_outputs'
    if not os.path.exists(output_path):
        os.makedirs(output_path)
    
    # centering X, Y coordinate
    for i in range(len(ad_list)):
        raw_loc = np.asarray(ad_list[i].obsm[raw_loc_key], dtype=np.float32)
        raw_loc[:,:2] = raw_loc[:,:2] - np.median(raw_loc[:,:2],axis=0)
        ad_list[i].obsm['spatial_pair'] = raw_loc
    
    start = time.time()
    print('Start alignment...')
    res = []
    for i in range(1,len(ad_list)):
        print(f'Alignment slice {i} to {i-1}')
        target_ind = i-1
        source_ind = i
        target_xy = ad_list[target_ind].obsm['spatial_pair'][:,:2]
        source_xy = ad_list[source_ind].obsm['spatial_pair'][:,:2]
        target_cluster = np.asarray(ad_list[target_ind].obs[cluster_key])
        source_cluster = np.asarray(ad_list[source_ind].obs[cluster_key])
        param = [i, target_xy, source_xy, target_cluster, source_cluster]
        r = align_pairwise(param, n_neighbors=n_neighbors, knn_exclude_cutoff=knn_exclude_cutoff,p=p,a=a,bound_alpha=bound_alpha, n_threads=n_threads,seed=seed)
        res.append(r)
    
    warp_info=np.array(res,dtype=np.float32)
    np.save(os.path.join(output_path,'warp_info.npy'),warp_info)
    
    warp_info = warp_info[:,:6]
    score = warp_info[:,5]
    print('All score:', str(score))
    print('Runtime: ' + str(time.time() - start),'s')

    ad_list[0].obsm[aligned_loc_key] = ad_list[0].obsm['spatial_pair']
    for i in range(1,len(ad_list)):
        target_ind = i-1
        source_ind = i
        target_loc = ad_list[target_ind].obsm['spatial_pair']
        source_loc = ad_list[source_ind].obsm['spatial_pair']
        target_cluster = np.asarray(ad_list[target_ind].obs[cluster_key])
        source_cluster = np.asarray(ad_list[source_ind].obs[cluster_key])
        old_source_loc = deepcopy(source_loc)
        for r in warp_info[:source_ind][::-1]:
            if r[1] == 1:
                loc_flip = deepcopy(old_source_loc)
                loc_flip[:,0] = -loc_flip[:,0]
                old_source_loc = loc_flip
            new_source_xy = []
            for _x,_y in old_source_loc[:,:2]:
                new_source_xy.append(rotate_loc(_x,_y,0,0,r[2]))
            new_source_xy = np.array(new_source_xy)
            new_source_loc = deepcopy(source_loc)
            new_source_loc[:,:2] = new_source_xy
            # translate loc
            new_source_loc[:,0] += r[3]
            new_source_loc[:,1] += r[4]
            old_source_loc = deepcopy(new_source_loc)
        ad_list[source_ind].obsm[aligned_loc_key] = new_source_loc

    for i in range(len(ad_list)):
        if isinstance (ad_list[i].obsm[raw_loc_key],pd.DataFrame):
            columns = ad_list[i].obsm[raw_loc_key].columns
        elif ad_list[i].obsm[raw_loc_key].shape[1] == 3:
            columns = ['X','Y','Z']
        else:
            columns = ['X','Y']
        aligned_loc = pd.DataFrame(ad_list[i].obsm[aligned_loc_key], columns=columns, index=ad_list[i].obs.index)
        ad_list[i].obsm[aligned_loc_key] = aligned_loc
    
    if write_loc_path is not None:
        coo = pd.DataFrame()
        for i in range(len(ad_list)):
            loc = ad_list[i].obsm[aligned_loc_key]
            coo = pd.concat([coo,loc],axis=0)
        coo.to_csv(write_loc_path)



import numpy as np
import pandas as pd
import random
import os 
import matplotlib
from sklearn.cluster import KMeans
from sklearn.metrics import davies_bouldin_score
import seaborn as sns
import squidpy as sq
from .graph import get_graph_inputs,get_graph_kernel,split_train_test_idx
from .utils import generate_celltype_ad_list,cal_celltype_weight,clustering
import math
import time
import numpy as np
from tqdm import tqdm

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.nn.parameter import Parameter
from torch.nn.modules.module import Module


class GraphConvolution(Module):
    def __init__(self, in_features, out_features, support, bias=True):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.support = support
        self.weight = Parameter(torch.FloatTensor(in_features * support, out_features))
        if bias:
            self.bias = Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, features, basis):
        supports = list()
        for i in range(self.support):
            supports.append(basis[i].matmul(features))
        supports = torch.cat(supports, dim=1)
        output = torch.spmm(supports, self.weight)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'
            
class Splane_GCN(nn.Module):
    def __init__(self,feature_dims,support,latent_dims=8,hidden_dims=64,dropout=0.8):
        super(Splane_GCN, self).__init__()
        self.feature_dims = feature_dims
        self.support = support
        self.latent_dims=latent_dims
        self.hidden_dims=hidden_dims
        self.dropout = dropout
        self.encode_gc1 = GraphConvolution(feature_dims, hidden_dims, support)
        self.encode_gc2 = GraphConvolution(hidden_dims, latent_dims, support)
        self.decode_gc1 = GraphConvolution(latent_dims, hidden_dims, support)
        self.decode_gc2 = GraphConvolution(hidden_dims, feature_dims, support)
        
        nn.init.kaiming_normal_(self.encode_gc1.weight)
        nn.init.xavier_uniform_(self.encode_gc2.weight)
        nn.init.kaiming_normal_(self.decode_gc1.weight)
        nn.init.xavier_uniform_(self.decode_gc2.weight)
        
    @staticmethod
    def l2_activate(x,dim):
        
        def scale(z):
            zmax = z.max(1, keepdims=True).values
            zmin = z.min(1, keepdims=True).values
            z_std = torch.nan_to_num(torch.div(z - zmin,(zmax - zmin)),0)
            return z_std
        
        x = scale(x)
        x = F.normalize(x, p=2, dim=1)
        return x
        
    def encode(self, x, adj):
        x = F.dropout(x, self.dropout, training=self.training)
        x = F.leaky_relu(self.encode_gc1(x, adj))
        x = F.dropout(x, self.dropout, training=self.training)
        x = self.encode_gc2(x, adj)
        return self.l2_activate(x, dim=1)
    
    def decode(self, x, adj):
        x = F.dropout(x, self.dropout, training=self.training)
        x = F.leaky_relu(self.decode_gc1(x, adj))
        x = F.dropout(x, self.dropout, training=self.training)
        x = self.decode_gc2(x, adj)
        return x

    def forward(self, x, adj):
        z = self.encode(x, adj)
        x_ = self.decode(z, adj)
        return z, x_
    
class Splane_Disc(nn.Module):
    def __init__(self,label,latent_dims=8,hidden_dims=64,dropout=0.5):
        super(Splane_Disc, self).__init__()
        self.latent_dims=latent_dims
        self.hidden_dims=hidden_dims
        self.dropout = dropout
        self.class_num = label.shape[1]
        self.disc = nn.Sequential(
            nn.Linear(latent_dims, hidden_dims),
            nn.LeakyReLU(),
            nn.BatchNorm1d(hidden_dims),
            nn.Linear(hidden_dims, hidden_dims),
            nn.LeakyReLU(),
            nn.BatchNorm1d(hidden_dims),
            nn.Dropout(dropout),
            nn.Linear(hidden_dims, self.class_num)
        )

    def forward(self, x):
        x = self.disc(x)
        y = F.softmax(x, dim=1)
        return y

class SplaneModel():
    def __init__(
        self,
        expr_ad_list,
        n_clusters,
        X,
        graph,
        support,
        slice_class_onehot,
        nb_mask,
        train_idx,
        test_idx,
        celltype_weights,
        morans_mean,
        lr,
        l1,
        l2,
        latent_dim,
        hidden_dims,
        gnn_dropout,
        use_gpu
    ):
        self.expr_ad_list = expr_ad_list
        self.model_g = Splane_GCN(X.shape[1],support,latent_dims=latent_dim,hidden_dims=hidden_dims,dropout=gnn_dropout)
        self.model_d = Splane_Disc(slice_class_onehot,latent_dims=latent_dim,hidden_dims=hidden_dims)
        self.graph = graph
        self.slice_class_onehot = slice_class_onehot
        self.nb_mask = nb_mask
        self.train_idx = train_idx
        self.test_idx = test_idx
        self.celltype_weights = torch.tensor(celltype_weights)
        self.morans_mean = morans_mean
        self.best_path = None
        self.cos_loss_obj = F.cosine_similarity
        self.d_loss_obj = F.cross_entropy
        self.n_clusters = n_clusters
        self.Cluster = KMeans(n_clusters=self.n_clusters,n_init=10,tol=1e-3,algorithm='full',max_iter=1000,random_state=42)
        self.optimizer_g = optim.RMSprop(self.model_g.parameters(),lr=lr)
        self.optimizer_d = optim.RMSprop(self.model_d.parameters(),lr=lr)
        self.l1 = l1
        self.l2 = l2
        if use_gpu:
            self.model_g = self.model_g.cuda()
            self.model_d = self.model_d.cuda()
            self.celltype_weights = self.celltype_weights.cuda()
    
    @staticmethod
    def kl_divergence(y_true, y_pred, dim=0):
        y_pred = torch.clip(y_pred, torch.finfo(torch.float32).eps)
        y_true = y_true.to(y_pred.dtype)
        y_true = torch.nan_to_num(torch.div(y_true, y_true.sum(dim, keepdims=True)),0)
        y_pred = torch.nan_to_num(torch.div(y_pred, y_pred.sum(dim, keepdims=True)),0)
        y_true = torch.clip(y_true, torch.finfo(torch.float32).eps, 1)
        y_pred = torch.clip(y_pred, torch.finfo(torch.float32).eps, 1)
        return torch.mul(y_true, torch.log(torch.nan_to_num(torch.div(y_true, y_pred)))).mean(dim)
        
    def train_model_g(self,d_l,simi_l):
        self.model_g.train()
        self.optimizer_g.zero_grad()
        encoded, decoded = self.model_g(self.graph[0],self.graph[1:])
        y_disc = self.model_d(encoded)
        d_loss = F.cross_entropy(self.slice_class_onehot, y_disc)
        decoded_mask = decoded[self.train_idx]
        x_mask = self.graph[0][self.train_idx]
        f_adj = torch.matmul(encoded, torch.transpose(encoded,0,1))
        simi_loss = -torch.mean(f_adj[self.nb_mask[0],self.nb_mask[1]]) + torch.mean(torch.abs(encoded[self.nb_mask[0]]-encoded[self.nb_mask[1]]))
        g_loss = torch.sum(self.celltype_weights*F.cosine_similarity(x_mask, decoded_mask,dim=0))+torch.sum(self.celltype_weights*self.kl_divergence(x_mask, decoded_mask, dim=0)) + simi_l*simi_loss
        
        regularization_params = torch.cat([
            torch.cat([x.view(-1) for x in self.model_g.encode_gc1.parameters()]),
            torch.cat([x.view(-1) for x in self.model_g.decode_gc1.parameters()])
        ])
        l1_regularization = self.l1 * torch.norm(regularization_params, 1)
        l2_regularization = self.l2 * torch.norm(regularization_params, 2)

        l1_l2_loss = l1_regularization + l2_regularization
        
        total_loss = g_loss - d_l*d_loss # + l1_l2_loss
        total_loss.backward()
        self.optimizer_g.step()
        return total_loss
    
    def train_model_d(self,):
        self.model_d.train()
        self.optimizer_d.zero_grad()
        encoded, decoded = self.model_g(self.graph[0],self.graph[1:])
        y_disc = self.model_d(encoded)
        d_loss = F.cross_entropy(self.slice_class_onehot, y_disc)
        d_loss.backward()
        self.optimizer_d.step()
        return d_loss
    
    def test_model(self,d_l,simi_l):
        self.model_g.eval()
        self.model_d.eval()
        encoded, decoded = self.model_g(self.graph[0],self.graph[1:])
        y_disc = self.model_d(encoded)
        d_loss = F.cross_entropy(self.slice_class_onehot, y_disc)
        decoded_mask = decoded[self.test_idx]
        x_mask = self.graph[0][self.test_idx]
        ll = torch.eq(torch.argmax(self.slice_class_onehot, -1), torch.argmax(y_disc, -1))
        accuarcy = ll.to(torch.float32).mean()
        f_adj = torch.matmul(encoded, torch.transpose(encoded,0,1))
        simi_loss = -torch.mean(f_adj[self.nb_mask[0],self.nb_mask[1]]) + torch.mean(torch.abs(encoded[self.nb_mask[0]]-encoded[self.nb_mask[1]]))
        g_loss = torch.sum(self.celltype_weights*F.cosine_similarity(x_mask, decoded_mask,dim=0))+torch.sum(self.celltype_weights*self.kl_divergence(x_mask, decoded_mask, dim=0)) + simi_l*simi_loss
        total_loss = g_loss - d_l*d_loss
        db_loss = clustering(self.Cluster, encoded.cpu().detach().numpy())
        return total_loss, g_loss, d_loss, accuarcy, simi_loss, db_loss, encoded, decoded
    
    def train(
        self,
        max_epochs=1000,
        convergence=0.0001,
        db_convergence=0,
        early_stop_epochs=10,
        d_l=0.2,
        simi_l=None,
        g_step = 1,
        d_step = 1,
        plot_step=5,
        save_path = 'Splane_models',
        prefix=None
    ):
        best_loss = np.inf
        best_db_loss = np.inf
        best_simi_loss = np.inf
        simi_l = 1/np.mean(self.morans_mean)

        if not os.path.exists(save_path):
            os.makedirs(save_path)
        early_stop_count = 0
        pbar = tqdm(range(max_epochs))
        for epoch in pbar:

            for _ in range(g_step):
                train_total_loss = self.train_model_g(d_l=d_l, simi_l=simi_l)
            for _ in range(d_step):
                self.train_model_d()

            if epoch % plot_step == 0:
                test_total_loss, test_g_loss, test_d_loss, test_acc, simi_loss, db_loss, encoded, decoded = self.test_model(d_l=d_l, simi_l=simi_l)
                current_loss = test_g_loss.cpu().detach().numpy()
                current_db_loss = db_loss
                if (best_loss - current_loss > convergence) & (best_db_loss - current_db_loss > db_convergence):
                    if best_loss > current_loss:
                        best_loss = current_loss
                    if best_db_loss > current_db_loss:
                        best_db_loss = current_db_loss
                    pbar.set_description("The best epoch {0} total loss={1:.3f} g loss={2:.3f} d loss={3:.3f} d acc={4:.3f} simi loss={5:.3f} db loss={6:.3f}".format(epoch, test_total_loss, test_g_loss, test_d_loss, test_acc, simi_loss, db_loss),refresh=True)
                    old_best_path = self.best_path
                    early_stop_count = 0
                    if prefix is not None:
                        self.best_path = os.path.join(save_path,prefix+'_'+f'Splane_weights_epoch{epoch}.h5')
                    else:
                        self.best_path = os.path.join(save_path,f'Splane_weights_epoch{epoch}.h5')
                    if old_best_path is not None:
                        if os.path.exists(old_best_path):
                            os.remove(old_best_path)
                    torch.save(self.model_g.state_dict(), self.best_path)
                else:
                    early_stop_count += 1
                

                # print("Epoch {} train g loss={} g loss={} d loss={} acc={} simi loss={} db loss={}".format(epoch, test_total_loss, test_g_loss, test_d_loss, test_acc, simi_loss, db_loss))
                if early_stop_count > early_stop_epochs:
                    print('Stop trainning because of loss convergence')
                    break
    
    def identify_spatial_domain(self,colors=None,key=None):
        if colors is None:
            if self.n_clusters > 10:
                colors = [matplotlib.colors.to_hex(c) for c in sns.color_palette('tab20',n_colors=self.n_clusters)]
            else:
                colors = [matplotlib.colors.to_hex(c) for c in sns.color_palette('tab10',n_colors=self.n_clusters)]
            color_map = pd.DataFrame(colors,index=np.arange(self.n_clusters),columns=['color'])
        if key is None:
            key = 'spatial_domain'
        # self.model_g(self.graph[0], self.graph[1:])
        self.model_g.load_state_dict(torch.load(self.best_path))
        self.model_g.eval()
        encoded, decoded = self.model_g(self.graph[0], self.graph[1:])
        clusters = self.Cluster.fit_predict(encoded.cpu().detach().numpy())
        loc_index = 0
        for i in range(len(self.expr_ad_list)):
            if key in self.expr_ad_list[i].obs.columns:
                self.expr_ad_list[i].obs = self.expr_ad_list[i].obs.drop(columns=key)
            self.expr_ad_list[i].obs[key] = clusters[loc_index:loc_index+self.expr_ad_list[i].shape[0]]
            self.expr_ad_list[i].obs[key] = pd.Categorical(self.expr_ad_list[i].obs[key])
            self.expr_ad_list[i].uns[f'{key}_colors'] = [color_map.loc[c,'color'] for c in self.expr_ad_list[i].obs['spatial_domain'].cat.categories]
            loc_index += self.expr_ad_list[i].shape[0]

def init_model(
    expr_ad_list:list,
    n_clusters:int,
    k:int=2,
    use_weight=True,
    train_prop:float=0.5,
    n_neighbors=6,
    min_prop=0.01,
    lr:float=3e-3,
    l1:float=0.01,
    l2:float=0.01,
    latent_dim:int=16,
    hidden_dims:int=64,
    gnn_dropout:float=0.8,
    simi_neighbor=1,
    use_gpu=None,
    seed=42
)->SplaneModel:
    """Initialize Splane model.
    
    Build the model then set the data and paratemters. 
    
    Args:
        expr_ad_list: A list of AnnData object of spatial transcriptomic data as the model input.
        n_clusters: The number of cluster of the model ouput. 
        k: The order of neighbors of a spot for graph construction.
        use_weight: If True, the cell type proportion of Moran was used as the weight of the loss.
        train_prop: The proportion of training set.
        n_neighbors: The number of neighbors for graph construction.
        lr: Learning rate of training.
        latent_dim: The dimension of latent features. It equal to the number of nodes of bottleneck layer.
        hidden_dims: The number of nodes of hidden layers.
        gnn_dropout: The dropout rate of GNN model.
        simi_neighbor: The order of neighbors used for similarity loss. If is None, It equal to the order used for constructed graph.
        seed: Random number seed.
    Returns:
        A ``DataFrame`` contained deconvoluted results. Each row representing a spot, and each column representing a cell type.
    """
    
    print('Setting global seed:', seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    
    if use_gpu is None:
        if torch.cuda.is_available():
            use_gpu = True
        else:
            use_gpu = False
    
    for expr_ad in expr_ad_list:
        if 'spatial_connectivities' not in expr_ad.obsp.keys():
            sq.gr.spatial_neighbors(expr_ad,coord_type='grid',n_neighs=n_neighbors)
    celltype_ad_list = generate_celltype_ad_list(expr_ad_list,min_prop=min_prop)
    celltype_weights,morans_mean = cal_celltype_weight(celltype_ad_list)
    kept_ind = celltype_weights > 0
    if not use_weight:
        celltype_weights = np.ones(len(celltype_weights))/len(celltype_weights)
    X,A,nb_mask,slice_class_onehot = get_graph_inputs(celltype_ad_list)
    X_filtered, graph, support = get_graph_kernel(X[:,kept_ind],A,k=k)
    celltype_weights = celltype_weights[kept_ind]
    morans_mean = morans_mean[kept_ind]
    if simi_neighbor == 1:
        nb_mask = nb_mask
    elif simi_neighbor == None:
        nb_mask = np.array(np.where(graph[-1].to_dense())!=0)
        nb_mask = nb_mask[:,nb_mask[0] != nb_mask[1]]
    else:
        raise ValueError('simi_neighbor must be 1 or None.')
    train_idx,test_idx = split_train_test_idx(X,train_prop=0.5)
    if use_gpu:
        for i in range(len(graph)):
            graph[i] = graph[i].cuda()
        slice_class_onehot = slice_class_onehot.cuda()
    return SplaneModel(
        expr_ad_list,
        n_clusters,
        X_filtered,
        graph,
        support,
        slice_class_onehot,
        nb_mask,
        train_idx,
        test_idx,
        celltype_weights,
        morans_mean,
        lr,
        l1,
        l2,
        latent_dim,
        hidden_dims,
        gnn_dropout,
        use_gpu
    )



import random
import anndata
import numpy as np
from scipy.sparse import coo_matrix
from sklearn.metrics import davies_bouldin_score

# From scanpy
def _morans_i_mtx(
    g_data: np.ndarray,
    g_indices: np.ndarray,
    g_indptr: np.ndarray,
    X: np.ndarray,
) -> np.ndarray:
    M, N = X.shape
    assert N == len(g_indptr) - 1
    W = g_data.sum()
    out = np.zeros(M, dtype=np.float_)
    for k in range(M):
        x = X[k, :]
        out[k] = _morans_i_vec_W(g_data, g_indices, g_indptr, x, W)
    return out

# From scanpy
def _morans_i_vec_W(
    g_data: np.ndarray,
    g_indices: np.ndarray,
    g_indptr: np.ndarray,
    x: np.ndarray,
    W: np.float_,
) -> float:
    z = x - x.mean()
    z2ss = (z * z).sum()
    N = len(x)
    inum = 0.0

    for i in range(N):
        s = slice(g_indptr[i], g_indptr[i + 1])
        i_indices = g_indices[s]
        i_data = g_data[s]
        inum += (i_data * z[i_indices]).sum() * z[i]

    return len(x) / W * inum / z2ss

def fill_low_prop(ad,min_prop):
    mtx = ad.X
    mtx[mtx < min_prop] = 0
    ad.X = mtx
    return ad

def cal_celltype_moran(ad):
    moran_vals = _morans_i_mtx(
        ad.obsp['spatial_connectivities'].data,
        ad.obsp['spatial_connectivities'].indices,
        ad.obsp['spatial_connectivities'].indptr,
        ad.X.T
    )
    ad.uns['moran_vals'] = np.nan_to_num(moran_vals)
    
def cal_celltype_weight(ad_list):
    print('Calculating cell type weights...')
    for ad in ad_list:
        cal_celltype_moran(ad)
    moran_min=-1
    morans = ad_list[0].uns['moran_vals'].copy()
    for i, ad in enumerate(ad_list[1:]):
        morans += ad.uns['moran_vals'].copy()
    morans_mean = morans/len(ad_list)
    celltype_weights = morans_mean/morans_mean.sum()
    return celltype_weights, morans_mean

def generate_celltype_ad_list(expr_ad_list,min_prop):
    celltype_ad_list = []
    for expr_ad in expr_ad_list:
        celltype_ad = anndata.AnnData(expr_ad.obs[[c for c in expr_ad.uns['celltypes']]])
        celltype_ad.obs = expr_ad.obs
        celltype_ad.obsm =expr_ad.obsm
        celltype_ad.obsp = expr_ad.obsp
        celltype_ad = fill_low_prop(celltype_ad,min_prop)
        celltype_ad_list.append(celltype_ad)
    return celltype_ad_list

def clustering(Cluster, feature):
    predict_labels = Cluster.fit_predict(feature)
    db = davies_bouldin_score(feature, predict_labels)
    return db
    
def split_ad(ad,by):
    ad_list = []
    for s in np.unique(ad.obs[by]):
        ad_split = ad[ad.obs[by] == s].copy()
        ad_list.append(ad_split)
    return ad_list




import pandas as pd
import numpy as np
import scanpy as sc
import anndata
from scipy.sparse import issparse,csr_matrix
from sklearn.preprocessing import normalize
from . import data_downsample
from . import data_augmentation
from . import spatial_simulation
import logging
# logging.basicConfig(level=print,
#                     format='%(asctime)s %(levelname)s %(message)s',
#                     datefmt='%m-%d %H:%M')
# logging.getLogger().setLevel(print)

from rpy2.robjects.packages import importr
import rpy2.robjects as robjects

def normalize_adata(ad,target_sum=None):
    ad_norm = sc.pp.normalize_total(ad,inplace=False,target_sum=1e4)
    ad_norm  = sc.pp.log1p(ad_norm['X'])
    # ad_norm  = sc.pp.scale(ad_norm)
    # ad_norm = normalize(ad_norm,axis=1)
    ad.layers['norm'] = ad_norm
    return ad

def normalize_mtx(mtx,target_sum):
    mtx = mtx[mtx.sum(1)!=0,:]
    mtx = np.nan_to_num(np.log1p((mtx.T*target_sum/mtx.sum(axis=1)).T))
    mtx = (mtx-mtx.min(axis=1,keepdims=True))/(mtx.max(axis=1,keepdims=True)-mtx.min(axis=1,keepdims=True))
    mtx = normalize(mtx,axis=1)
    return mtx
    
# 计算单细胞亚群差异基因
def find_sc_markers(sc_ad, celltype_key, layer='norm', deg_method=None, log2fc_min=0.5, pval_cutoff=0.01, n_top_markers=200, pct_diff=None, pct_min=0.1):
    print('### Finding marker genes...')
    
    # filter celltype contain only one sample.
    filtered_celltypes = list(sc_ad.obs[celltype_key].value_counts()[(sc_ad.obs[celltype_key].value_counts() == 1).values].index)
    if len(filtered_celltypes) > 0:
        sc_ad = sc_ad[sc_ad.obs[~(sc_ad.obs[celltype_key].isin(filtered_celltypes))].index,:].copy()
        print(f'### Filter cluster contain only one sample: {filtered_celltypes}')

    sc.tl.rank_genes_groups(sc_ad, groupby=celltype_key, pts=True, layer=layer, use_raw=False, method=deg_method)
    marker_genes_dfs = []
    for c in np.unique(sc_ad.obs[celltype_key]):
        tmp_marker_gene_df = sc.get.rank_genes_groups_df(sc_ad, group=c, pval_cutoff=pval_cutoff, log2fc_min=log2fc_min)
        if (tmp_marker_gene_df.empty is not True):
            tmp_marker_gene_df.index = tmp_marker_gene_df.names
            tmp_marker_gene_df.loc[:,celltype_key] = c
            if pct_diff is not None:
                pct_diff_genes = sc_ad.var_names[np.where((sc_ad.uns['rank_genes_groups']['pts'][c]-sc_ad.uns['rank_genes_groups']['pts_rest'][c]) > pct_diff)]
                tmp_marker_gene_df = tmp_marker_gene_df.loc[np.intersect1d(pct_diff_genes, tmp_marker_gene_df.index),:]
            if pct_min is not None:
                # pct_min_genes = sc_ad.var_names[np.where((sc_ad.uns['rank_genes_groups']['pts'][c]) > pct_min)]
                tmp_marker_gene_df = tmp_marker_gene_df[tmp_marker_gene_df['pct_nz_group'] > pct_min]
            if n_top_markers is not None:
                tmp_marker_gene_df = tmp_marker_gene_df.sort_values('logfoldchanges',ascending=False)
                tmp_marker_gene_df = tmp_marker_gene_df.iloc[:n_top_markers,:]
            marker_genes_dfs.append(tmp_marker_gene_df)
    marker_gene_df = pd.concat(marker_genes_dfs,axis=0)
    print(marker_gene_df[celltype_key].value_counts())
    all_marker_genes = np.unique(marker_gene_df.names)
    return all_marker_genes

# 计算空间HVG
def find_st_hvg(st_ad,n_top_hvg=None):
    print('### Finding HVG in spatial...')
    sc.pp.highly_variable_genes(st_ad,n_top_genes=n_top_hvg,flavor='seurat_v3')
    return st_ad.var_names[st_ad.var['highly_variable'] == True]

# 
def filter_model_genes(
    sc_ad, 
    st_ad, 
    celltype_key=None, 
    used_genes=None,
    sc_genes=None,
    st_genes=None,
    layer='norm',
    deg_method=None,
    log2fc_min=0.5, 
    pval_cutoff=0.01, 
    n_top_markers:int=200, 
    n_top_hvg=None,
    pct_diff=None, 
    pct_min=0.1,
):
    overlaped_genes = np.intersect1d(sc_ad.var_names,st_ad.var_names)
    sc_ad = sc_ad[:,overlaped_genes].copy()
    st_ad = st_ad[:,overlaped_genes].copy()
    if used_genes is None:
        if st_genes is None:
            if n_top_hvg is None:
                st_genes = st_ad.var_names
            else:
                st_genes = find_st_hvg(st_ad, n_top_hvg)
        if sc_genes is None:
            sc_ad = sc_ad[:, st_genes].copy()
            sc_genes = find_sc_markers(sc_ad, celltype_key, layer, deg_method, log2fc_min, pval_cutoff, n_top_markers, pct_diff, pct_min)
        used_genes = np.intersect1d(sc_genes,st_genes)
    sc_ad = sc_ad[:,used_genes].copy()
    st_ad = st_ad[:,used_genes].copy()
    sc.pp.filter_cells(sc_ad,min_genes=1)
    sc.pp.filter_cells(st_ad,min_genes=1)
    print(f'### Used gene numbers: {len(used_genes)}')
    return sc_ad, st_ad

def check_data_type(ad):
    if issparse(ad.X):
        ad.X = ad.X.toarray()
    if ad.X.dtype != np.float32:
        ad.X =ad.X.astype(np.float32)
    return ad

def generate_sm_adata(sc_ad,num_sample,celltype_key,n_threads,cell_counts,clusters_mean,cells_mean,cells_min,cells_max):
    sm_data,sm_labels = spatial_simulation.generate_simulation_data(sc_ad,num_sample=num_sample,celltype_key=celltype_key,downsample_fraction=None,data_augmentation=False,n_cpus=n_threads,cell_counts=cell_counts,clusters_mean=clusters_mean,cells_mean=cells_mean,cells_min=cells_min,cells_max=cells_max)
    sm_data_mtx = csr_matrix(sm_data)
    sm_ad = anndata.AnnData(sm_data_mtx)
    sm_ad.var.index = sc_ad.var_names
    sm_labels = (sm_labels.T/sm_labels.sum(axis=1)).T
    sm_ad.obsm['label'] = pd.DataFrame(sm_labels,columns=np.array(sc_ad.obs[celltype_key].value_counts().index.values),index=sm_ad.obs_names)
    return sm_ad

def downsample_sm_spot_counts(sm_ad,st_ad,n_threads):
    fitdistrplus = importr('fitdistrplus')
    lib_sizes = robjects.FloatVector(np.array(st_ad.X.sum(1)).reshape(-1))
    res = fitdistrplus.fitdist(lib_sizes,'lnorm')
    loc = res[0][0]
    scale = res[0][1]

    sm_mtx_count = sm_ad.X.toarray()
    sample_cell_counts = np.random.lognormal(loc,scale,sm_ad.shape[0])
    sm_mtx_count_lb = data_downsample.downsample_matrix_by_cell(sm_mtx_count,sample_cell_counts.astype(np.int64), n_cpus=n_threads, numba_end=False)
    sm_ad.X = csr_matrix(sm_mtx_count_lb)
    
def split_shuffle_data(X,Y,shuffle=True,proportion=0.8):
    if shuffle:
        reind = np.random.permutation(len(X))
        X = X[reind]
        Y = Y[reind]
    X_train = X[:int(len(X)*proportion)]
    Y_train = Y[:int(len(Y)*proportion)]
    X_test = X[int(len(X)*proportion):]
    Y_test = Y[int(len(Y)*proportion):]
    return X_train,Y_train,X_test,Y_test



import numpy as np
import pandas as pd
import numba as nb
from numba import jit
import collections
import random
from .data_downsample import downsample_cell,downsample_matrix_by_cell
from .data_augmentation import random_augment,random_augmentation_cell
import logging
# logging.basicConfig(level=print,
#                     format='%(asctime)s %(levelname)s %(message)s',
#                     datefmt='%m-%d %H:%M')
# logging.getLogger().setLevel(print)

# 汇总每个spot的细胞数，统计细胞数的分布
def count_cell_counts(cell_counts):
    cell_counts = np.array(cell_counts.values,dtype=int).reshape(-1)
    counts_list = np.array(np.histogram(cell_counts,range=[0,np.max(cell_counts)+1],bins=np.max(cell_counts)+1)[0],dtype=int)
    counts_index = np.array((np.histogram(cell_counts,range=[0,np.max(cell_counts)+1],bins=np.max(cell_counts)+1)[1][:-1]),dtype=int)
    counts_df = pd.DataFrame(counts_list,index=counts_index,columns=['count'],dtype=np.int32)
    counts_df = counts_df[(counts_df['count'] != 0) & (counts_df.index != 0)]
    count_sum = 0
    for i in np.arange(len(counts_df)):
        count_sum += counts_df.iloc[i].values
        if count_sum > counts_df.values.sum()*0.99:
            counts_df_filtered = counts_df.iloc[:i+1,:]
            break
    return counts_df_filtered


@nb.njit
def numba_set_seed(seed):
    np.random.seed(seed)
    random.seed(seed)

# 对某个axis调用numpy函数(numba版本)
@nb.njit
def np_apply_along_axis(func1d, axis, arr):
    assert arr.ndim == 2
    assert axis in [0, 1]
    if axis == 0:
        result = np.empty(arr.shape[1], dtype=arr.dtype)
        for i in range(len(result)):
            result[i] = func1d(arr[:, i])
    else:
        result = np.empty(arr.shape[0], dtype=arr.dtype)
        for i in range(len(result)):
            result[i] = func1d(arr[i, :])
    return result

# 对某个axis计算均值(numba版本)
@nb.njit
def np_mean(array, axis):
    return np_apply_along_axis(np.mean, axis, array)

# 对某个axis计算加和(numba版本)
@nb.njit
def np_sum(array, axis):
    return np_apply_along_axis(np.sum, axis, array)

# 根据参数采样单细胞数据，生成模拟spot(numba版本)
@jit(nopython=True,parallel=True)
def sample_cell(param_list,cluster_p,clusters,cluster_id,sample_exp,sample_cluster,cell_p_balanced,downsample_fraction=None,data_augmentation=True,max_rate=0.8,max_val=0.8,kth=0.2):
    exp = np.empty((len(param_list), sample_exp.shape[1]),dtype=np.float32)
    density = np.empty((len(param_list), sample_cluster.shape[1]),dtype=np.float32)

    for i in nb.prange(len(param_list)):
        params = param_list[i]
        num_cell = params[0]
        num_cluster = params[1]
        used_clusters = clusters[np.searchsorted(np.cumsum(cluster_p), np.random.rand(num_cluster), side="right")]
        cluster_mask=np.array([False]*len(cluster_id))
        for c in used_clusters:
            cluster_mask = (cluster_id==c)|(cluster_mask)
        used_cell_ind = np.where(cluster_mask)[0]
        used_cell_p = cell_p_balanced[cluster_mask]
        used_cell_p = used_cell_p/used_cell_p.sum()
        sampled_cells = used_cell_ind[np.searchsorted(np.cumsum(used_cell_p), np.random.rand(num_cell), side="right")]
        combined_exp = np_sum(sample_exp[sampled_cells,:],axis=0).astype(np.float32)
        if data_augmentation:
            combined_exp = random_augmentation_cell(combined_exp,max_rate=max_rate,max_val=max_val,kth=kth)
        if downsample_fraction is not None:
            combined_exp = downsample_cell(combined_exp, downsample_fraction)
        combined_clusters = np_sum(sample_cluster[cluster_id[sampled_cells]],axis=0).astype(np.float32)
        exp[i,:] = combined_exp
        density[i,:] = combined_clusters
    return exp,density

def init_sample_prob(sc_ad,celltype_key):
    print('### Initializing sample probability')
    sc_ad.uns['celltype2num'] = pd.DataFrame(
        np.arange(len(sc_ad.obs[celltype_key].value_counts())).T,
        index=sc_ad.obs[celltype_key].value_counts().index.values,
        columns=['celltype_num']
    )
    sc_ad.obs['celltype_num'] = [sc_ad.uns['celltype2num'].loc[c,'celltype_num'] for c in sc_ad.obs[celltype_key]]
    cluster_p_unbalance = sc_ad.obs['celltype_num'].value_counts()/sc_ad.obs['celltype_num'].value_counts().sum()
    cluster_p_sqrt = np.sqrt(sc_ad.obs['celltype_num'].value_counts())/np.sqrt(sc_ad.obs['celltype_num'].value_counts()).sum()
    cluster_p_balance = pd.Series(
        np.ones(len(sc_ad.obs['celltype_num'].value_counts()))/len(sc_ad.obs['celltype_num'].value_counts()), 
        index=sc_ad.obs['celltype_num'].value_counts().index
    )
#     cluster_p_balance = np.ones(len(sc_ad.obs['celltype_num'].value_counts()))/len(sc_ad.obs['celltype_num'].value_counts())
    cell_p_balanced = [1/cluster_p_unbalance[c] for c in sc_ad.obs['celltype_num']]
    cell_p_balanced = np.array(cell_p_balanced)/np.array(cell_p_balanced).sum()
    sc_ad.obs['cell_p_balanced'] = cell_p_balanced
    sc_ad.uns['cluster_p_balance'] = cluster_p_balance
    sc_ad.uns['cluster_p_sqrt'] = cluster_p_sqrt
    sc_ad.uns['cluster_p_unbalance'] = cluster_p_unbalance
    return sc_ad

# 将表达矩阵转化成array
def generate_sample_array(sc_ad, used_genes):
    if used_genes is not None:
        sc_df = sc_ad.to_df().loc[:,used_genes]
    else:
        sc_df = sc_ad.to_df()
    return sc_df.values

# 从均匀分布中获取每个spot采样的细胞数和细胞类型数
def get_param_from_uniform(num_sample,cells_min=None,cells_max=None,clusters_min=None,clusters_max=None):

    cell_count = np.asarray(np.ceil(np.random.uniform(int(cells_min),int(cells_max),size=num_sample)),dtype=int)
    cluster_count = np.asarray(np.ceil(np.clip(np.random.uniform(clusters_min,clusters_max,size=num_sample),1,cell_count)),dtype=int)
    return cell_count, cluster_count

# 从高斯分布中获取每个spot采样的细胞数和细胞类型数
def get_param_from_gaussian(num_sample,cells_min=None,cells_max=None,cells_mean=None,cells_std=None,clusters_mean=None,clusters_std=None):

    cell_count = np.asarray(np.ceil(np.clip(np.random.normal(cells_mean,cells_std,size=num_sample),int(cells_min),int(cells_max))),dtype=int)
    cluster_count = np.asarray(np.ceil(np.clip(np.random.normal(clusters_mean,clusters_std,size=num_sample),1,cell_count)),dtype=int)
    return cell_count,cluster_count

# 从用空间数据估计的cell counts中获取每个spot采样的细胞数和细胞类型数
def get_param_from_cell_counts(
    num_sample,
    cell_counts,
    cluster_sample_mode='gaussian',
    cells_min=None,cells_max=None,
    cells_mean=None,cells_std=None,
    clusters_mean=None,clusters_std=None,
    clusters_min=None,clusters_max=None
):
    cell_count = np.asarray(np.ceil(np.clip(np.random.normal(cells_mean,cells_std,size=num_sample),int(cells_min),int(cells_max))),dtype=int)
    if cluster_sample_mode == 'gaussian':
        cluster_count = np.asarray(np.ceil(np.clip(np.random.normal(clusters_mean,clusters_std,size=num_sample),1,cell_count)),dtype=int)
    elif cluster_sample_mode == 'uniform':
        cluster_count = np.asarray(np.ceil(np.clip(np.random.uniform(clusters_min,clusters_max,size=num_sample),1,cell_count)),dtype=int)
    else:
        raise TypeError('Not correct sample method.')
    return cell_count,cluster_count

# 获取每个cluster的采样概率
def get_cluster_sample_prob(sc_ad,mode):
    if mode == 'unbalance':
        cluster_p = sc_ad.uns['cluster_p_unbalance'].values
    elif mode == 'balance':
        cluster_p = sc_ad.uns['cluster_p_balance'].values
    elif mode == 'sqrt':
        cluster_p = sc_ad.uns['cluster_p_sqrt'].values
    else:
        raise TypeError('Balance argument must be one of [ None, banlance, sqrt ].')
    return cluster_p

def cal_downsample_fraction(sc_ad,st_ad,celltype_key=None):
    st_counts_median = np.median(st_ad.X.sum(axis=1))
    simulated_st_data, simulated_st_labels = generate_simulation_data(sc_ad,num_sample=10000,celltype_key=celltype_key,balance_mode=['unbalance'])
    simulated_st_counts_median = np.median(simulated_st_data.sum(axis=1))
    if st_counts_median < simulated_st_counts_median:
        fraction = st_counts_median / simulated_st_counts_median
        print(f'### Simulated data downsample fraction: {fraction}')
        return fraction
    else:
        return None

# 生成模拟数据
def generate_simulation_data(
    sc_ad,
    celltype_key,
    num_sample: int, 
    used_genes=None,
    balance_mode=['unbalance','sqrt','balance'],
    cell_sample_method='gaussian',
    cluster_sample_method='gaussian',
    cell_counts=None,
    downsample_fraction=None,
    data_augmentation=True,
    max_rate=0.8,max_val=0.8,kth=0.2,
    cells_min=1,cells_max=20,
    cells_mean=10,cells_std=5,
    clusters_mean=None,clusters_std=None,
    clusters_min=None,clusters_max=None,
    n_cpus=None
):
    if not 'cluster_p_unbalance' in sc_ad.uns:
        sc_ad = init_sample_prob(sc_ad,celltype_key)
    num_sample_per_mode = num_sample//len(balance_mode)
    cluster_ordered = np.array(sc_ad.obs['celltype_num'].value_counts().index)
#     print(cluster_ordered)
    cluster_num = len(cluster_ordered)
#     print(cluster_num)
    cluster_id = sc_ad.obs['celltype_num'].values
#     print(cluster_id)
    cluster_mask = np.eye(cluster_num)
#     print(cluster_mask)
    
    if cell_counts is not None:
        cells_mean = np.mean(np.sort(cell_counts)[int(len(cell_counts)*0.05):int(len(cell_counts)*0.95)])
        cells_std = np.std(np.sort(cell_counts)[int(len(cell_counts)*0.05):int(len(cell_counts)*0.95)])
        cells_min = int(np.min(np.sort(cell_counts)[int(len(cell_counts)*0.05):int(len(cell_counts)*0.95)]))
        cells_max = int(np.max(np.sort(cell_counts)[int(len(cell_counts)*0.05):int(len(cell_counts)*0.95)]))
    if clusters_mean is None:
        clusters_mean = cells_mean/2
    if clusters_std is None:
        clusters_std = cells_std/2
    if clusters_min is None:
        clusters_min = cells_min
    if clusters_max is None:
        clusters_max = np.min((cells_max//2,cluster_num))
            
    if cell_counts is not None:
        cell_counts, cluster_count = get_param_from_cell_counts(num_sample_per_mode,cell_counts,cluster_sample_method,cells_mean=cells_mean,cells_std=cells_std,cells_max=cells_max,cells_min=cells_min,clusters_mean=clusters_mean,clusters_std=clusters_std,clusters_min=clusters_min,clusters_max=clusters_max)
    elif cell_sample_method == 'gaussian':
        cell_counts, cluster_count = get_param_from_gaussian(num_sample_per_mode,cells_mean=cells_mean,cells_std=cells_std,cells_max=cells_max,cells_min=cells_min,clusters_mean=clusters_mean,clusters_std=clusters_std)
    elif cell_sample_method == 'uniform':
        cell_counts, cluster_count = get_param_from_uniform(num_sample_per_mode,cells_max=cells_max,cells_min=cells_min,clusters_min=clusters_min,clusters_max=clusters_max)
    else:
        raise TypeError('Not correct sample method.')
    params = np.array(list(zip(cell_counts, cluster_count)))

    sample_data_list = []
    sample_labels_list = []
    for b in balance_mode:
        print(f'### Genetating simulated spatial data using scRNA data with mode: {b}')
        cluster_p = get_cluster_sample_prob(sc_ad,b)
        if downsample_fraction is not None:
            if downsample_fraction > 0.035:
                sample_data,sample_labels = sample_cell(
                    param_list=params,
                    cluster_p=cluster_p,
                    clusters=cluster_ordered,
                    cluster_id=cluster_id,
                    sample_exp=generate_sample_array(sc_ad,used_genes),
                    sample_cluster=cluster_mask,
                    cell_p_balanced=sc_ad.obs['cell_p_balanced'].values,
                    downsample_fraction=downsample_fraction,
                    data_augmentation=data_augmentation,max_rate=max_rate,max_val=max_val,kth=kth,
                )
            else:
                sample_data,sample_labels = sample_cell(
                    param_list=params,
                    cluster_p=cluster_p,
                    clusters=cluster_ordered,
                    cluster_id=cluster_id,
                    sample_exp=generate_sample_array(sc_ad,used_genes),
                    sample_cluster=cluster_mask,
                    cell_p_balanced=sc_ad.obs['cell_p_balanced'].values,
                    data_augmentation=data_augmentation,max_rate=max_rate,max_val=max_val,kth=kth,
                )
                # logging.warning('### Downsample data with python backend')
                sample_data = downsample_matrix_by_cell(sample_data, downsample_fraction, n_cpus=n_cpus, numba_end=False)
        else:
            sample_data,sample_labels = sample_cell(
                param_list=params,
                cluster_p=cluster_p,
                clusters=cluster_ordered,
                cluster_id=cluster_id,
                sample_exp=generate_sample_array(sc_ad,used_genes),
                sample_cluster=cluster_mask,
                cell_p_balanced=sc_ad.obs['cell_p_balanced'].values,
                data_augmentation=data_augmentation,max_rate=max_rate,max_val=max_val,kth=kth,
            )
#         if data_augmentation:
#             sample_data = random_augment(sample_data)
        sample_data_list.append(sample_data)
        sample_labels_list.append(sample_labels)
    return np.concatenate(sample_data_list), np.concatenate(sample_labels_list)

@jit(nopython=True,parallel=True)
def sample_cell_exp(cell_counts,sample_exp,cell_p,downsample_fraction=None,data_augmentation=True,max_rate=0.8,max_val=0.8,kth=0.2):
    exp = np.empty((len(cell_counts), sample_exp.shape[1]),dtype=np.float32)
    ind = np.zeros((len(cell_counts), np.max(cell_counts)),dtype=np.int32)
    cell_ind = np.arange(sample_exp.shape[0])
    for i in nb.prange(len(cell_counts)):
        num_cell = cell_counts[i]
        sampled_cells=cell_ind[np.searchsorted(np.cumsum(cell_p), np.random.rand(num_cell), side="right")]
        combined_exp=np_sum(sample_exp[sampled_cells,:],axis=0).astype(np.float64)
#         print(combined_exp.dtype)
        if downsample_fraction is not None:
            combined_exp = downsample_cell(combined_exp, downsample_fraction)
        if data_augmentation:
            combined_exp = random_augmentation_cell(combined_exp,max_rate=max_rate,max_val=max_val,kth=kth)
        exp[i,:] = combined_exp
        ind[i,:cell_counts[i]] = sampled_cells + 1
    return exp,ind

def generate_simulation_st_data(
    st_ad,
    num_sample: int, 
    used_genes=None,
    balance_mode=['unbalance'],
    cell_sample_method='gaussian',
    cell_counts=None,
    downsample_fraction=None,
    data_augmentation=True,
    max_rate=0.8,max_val=0.8,kth=0.2,
    cells_min=1,cells_max=10,
    cells_mean=5,cells_std=3,
):
    print('### Genetating simulated data using spatial data')
    cell_p = np.ones(len(st_ad))/len(st_ad)
    if cell_counts is not None:
        cells_mean = np.mean(np.sort(cell_counts)[int(len(cell_counts)*0.05):int(len(cell_counts)*0.95)])
        cells_std = np.std(np.sort(cell_counts)[int(len(cell_counts)*0.05):int(len(cell_counts)*0.95)])
        cells_min = int(np.min(np.sort(cell_counts)[int(len(cell_counts)*0.05):int(len(cell_counts)*0.95)]))
        cells_max = int(np.max(np.sort(cell_counts)[int(len(cell_counts)*0.05):int(len(cell_counts)*0.95)]))
    elif cell_sample_method == 'gaussian':
        cell_counts = np.asarray(np.ceil(np.clip(np.random.normal(cells_mean,cells_std,size=num_sample),int(cells_min),int(cells_max))),dtype=int)
    elif cell_sample_method == 'uniform':
        cell_counts = np.asarray(np.ceil(np.random.uniform(int(cells_min),int(cells_max),size=num_sample)),dtype=int)
    else:
        raise TypeError('Not correct sample method.')

    sample_data,sample_ind = sample_cell_exp(
        cell_counts=cell_counts,
        sample_exp=generate_sample_array(st_ad,used_genes),
        cell_p=cell_p,
        downsample_fraction=downsample_fraction,
        data_augmentation=data_augmentation,max_rate=max_rate,max_val=max_val,kth=kth,
    )
#     if data_augmentation:
#         sample_data = random_augment(sample_data)
    return sample_data,sample_ind



import numba
import numpy as np
import multiprocessing as mp
from functools import partial
import random

# Cite from https://github.com/numba/numba-examples
@numba.jit(nopython=True, parallel=True)
def get_bin_edges(a, bins):
    bin_edges = np.zeros((bins+1,), dtype=np.float32)
    a_min = a.min()
    a_max = a.max()
    delta = (a_max - a_min) / bins
    for i in numba.prange(bin_edges.shape[0]):
        bin_edges[i] = a_min + i * delta

    bin_edges[-1] = a_max  # Avoid roundoff error on last point
    return bin_edges

# Modified from https://github.com/numba/numba-examples
@numba.jit(nopython=True, parallel=False)
def compute_bin(x, bin_edges):
    # assuming uniform bins for now
    n = bin_edges.shape[0] - 1
    a_max = bin_edges[-1]
    # special case to mirror NumPy behavior for last bin
    if x == a_max:
        return n - 1 # a_max always in last bin
    bin = np.searchsorted(bin_edges, x)-1
    if bin < 0 or bin >= n:
        return None
    else:
        return bin

# Modified from https://github.com/numba/numba-examples
@numba.jit(nopython=True, parallel=False)
def numba_histogram(a, bin_edges):
    hist = np.zeros((bin_edges.shape[0] - 1,), dtype=np.intp)
    for x in a.flat:
        bin = compute_bin(x, bin_edges)
        if bin is not None:
            hist[int(bin)] += 1
    return hist, bin_edges


# Modified from https://rdrr.io/bioc/scRecover/src/R/countsSampling.R
# Downsample cell reads to a fraction
@numba.jit(nopython=True, parallel=True)
def downsample_cell(cell_counts,fraction):
    n = np.floor(np.sum(cell_counts) * fraction)
    readsGet = np.sort(np.random.choice(np.arange(np.sum(cell_counts)), np.intp(n), replace=False))
    cumCounts = np.concatenate((np.array([0]),np.cumsum(cell_counts)))
    counts_new = numba_histogram(readsGet,cumCounts)[0]
    counts_new = counts_new.astype(np.float32)
    return counts_new

def downsample_cell_python(cell_counts,fraction):
    n = np.floor(np.sum(cell_counts) * fraction)
    readsGet = np.sort(random.sample(range(np.intp(np.sum(cell_counts))), np.intp(n)))
    cumCounts = np.concatenate((np.array([0]),np.cumsum(cell_counts)))
    counts_new = numba_histogram(readsGet,cumCounts)[0]
    counts_new = counts_new.astype(np.float32)
    return counts_new

@numba.jit(nopython=True, parallel=True)
def downsample_per_cell(cell_counts,new_cell_counts):
    n = new_cell_counts
    if n < np.sum(cell_counts):
        readsGet = np.sort(np.random.choice(np.arange(np.sum(cell_counts)), np.intp(n), replace=False))
        cumCounts = np.concatenate((np.array([0]),np.cumsum(cell_counts)))
        counts_new = numba_histogram(readsGet,cumCounts)[0]
        counts_new = counts_new.astype(np.float32)
        return counts_new
    else:
        return cell_counts.astype(np.float32)

def downsample_per_cell_python(param):
    cell_counts,new_cell_counts = param[0],param[1]
    n = new_cell_counts
    if n < np.sum(cell_counts):
        readsGet = np.sort(random.sample(range(np.intp(np.sum(cell_counts))), np.intp(n)))
        cumCounts = np.concatenate((np.array([0]),np.cumsum(cell_counts)))
        counts_new = numba_histogram(readsGet,cumCounts)[0]
        counts_new = counts_new.astype(np.float32)
        return counts_new
    else:
        return cell_counts.astype(np.float32)

def downsample_matrix_by_cell(matrix,per_cell_counts,n_cpus=None,numba_end=True):
    if numba_end:
        downsample_func = downsample_per_cell
    else:
        downsample_func = downsample_per_cell_python
    if n_cpus is not None:
        with mp.Pool(n_cpus) as p:
            matrix_ds = p.map(downsample_func, zip(matrix,per_cell_counts))
    else:
        matrix_ds = [downsample_func(c,per_cell_counts[i]) for i,c in enumerate(matrix)]
    return np.array(matrix_ds)

# ps. slow speed.
def downsample_matrix_total(matrix,fraction):
    matrix_flat = matrix.reshape(-1)
    matrix_flat_ds = downsample_cell(matrix_flat,fraction)
    matrix_ds = matrix_flat_ds.reshape(matrix.shape)
    return matrix_ds




from __future__ import print_function

import scipy.sparse as sp
import numpy as np
from scipy.sparse.linalg.eigen.arpack import eigsh, ArpackNoConvergence


def encode_onehot(labels):
    classes = set(labels)
    classes_dict = {c: np.identity(len(classes))[i, :] for i, c in enumerate(classes)}
    labels_onehot = np.array(list(map(classes_dict.get, labels)), dtype=np.int32)
    return labels_onehot


def load_data(path="data/cora/", dataset="cora"):
    """Load citation network dataset (cora only for now)"""
    print('Loading {} dataset...'.format(dataset))

    idx_features_labels = np.genfromtxt("{}{}.content".format(path, dataset), dtype=np.dtype(str))
    features = sp.csr_matrix(idx_features_labels[:, 1:-1], dtype=np.float32)
    labels = encode_onehot(idx_features_labels[:, -1])

    # build graph
    idx = np.array(idx_features_labels[:, 0], dtype=np.int32)
    idx_map = {j: i for i, j in enumerate(idx)}
    edges_unordered = np.genfromtxt("{}{}.cites".format(path, dataset), dtype=np.int32)
    edges = np.array(list(map(idx_map.get, edges_unordered.flatten())),
                     dtype=np.int32).reshape(edges_unordered.shape)
    adj = sp.coo_matrix((np.ones(edges.shape[0]), (edges[:, 0], edges[:, 1])),
                        shape=(labels.shape[0], labels.shape[0]), dtype=np.float32)

    # build symmetric adjacency matrix
    adj = adj + adj.T.multiply(adj.T > adj) - adj.multiply(adj.T > adj)

    print('Dataset has {} nodes, {} edges, {} features.'.format(adj.shape[0], edges.shape[0], features.shape[1]))

    return features.todense(), adj, labels


def normalize_adj(adj, symmetric=True):
    if symmetric:
        d = sp.diags(np.power(np.array(adj.sum(1)), -0.5).flatten(), 0)
        a_norm = adj.dot(d).transpose().dot(d).tocsr()
    else:
        d = sp.diags(np.power(np.array(adj.sum(1)), -1).flatten(), 0)
        a_norm = d.dot(adj).tocsr()
    return a_norm


def preprocess_adj(adj, symmetric=True):
    adj = adj + sp.eye(adj.shape[0])
    adj = normalize_adj(adj, symmetric)
    return adj


def sample_mask(idx, l):
    mask = np.zeros(l)
    mask[idx] = 1
    return np.array(mask, dtype=np.bool)


def get_splits(y):
    idx_train = range(140)
    idx_val = range(200, 500)
    idx_test = range(500, 1500)
    y_train = np.zeros(y.shape, dtype=np.int32)
    y_val = np.zeros(y.shape, dtype=np.int32)
    y_test = np.zeros(y.shape, dtype=np.int32)
    y_train[idx_train] = y[idx_train]
    y_val[idx_val] = y[idx_val]
    y_test[idx_test] = y[idx_test]
    train_mask = sample_mask(idx_train, y.shape[0])
    return y_train, y_val, y_test, idx_train, idx_val, idx_test, train_mask


def categorical_crossentropy(preds, labels):
    return np.mean(-np.log(np.extract(labels, preds)))


def accuracy(preds, labels):
    return np.mean(np.equal(np.argmax(labels, 1), np.argmax(preds, 1)))


def evaluate_preds(preds, labels, indices):

    split_loss = list()
    split_acc = list()

    for y_split, idx_split in zip(labels, indices):
        split_loss.append(categorical_crossentropy(preds[idx_split], y_split[idx_split]))
        split_acc.append(accuracy(preds[idx_split], y_split[idx_split]))

    return split_loss, split_acc


def normalized_laplacian(adj, symmetric=True):
    adj_normalized = normalize_adj(adj, symmetric)
    laplacian = sp.eye(adj.shape[0]) - adj_normalized
    return laplacian


def rescale_laplacian(laplacian):
    try:
        print('Calculating largest eigenvalue of normalized graph Laplacian...')
        largest_eigval = eigsh(laplacian, 1, which='LM', return_eigenvectors=False)[0]
    except ArpackNoConvergence:
        print('Eigenvalue calculation did not converge! Using largest_eigval=2 instead.')
        largest_eigval = 2

    scaled_laplacian = (2. / largest_eigval) * laplacian - sp.eye(laplacian.shape[0])
    return scaled_laplacian


def chebyshev_polynomial(X, k):
    """Calculate Chebyshev polynomials up to order k. Return a list of sparse matrices."""
    print("Calculating Chebyshev polynomials up to order {}...".format(k))

    T_k = list()
    T_k.append(sp.eye(X.shape[0]).tocsr())
    T_k.append(X)

    def chebyshev_recurrence(T_k_minus_one, T_k_minus_two, X):
        X_ = sp.csr_matrix(X, copy=True)
        return 2 * X_.dot(T_k_minus_one) - T_k_minus_two

    for i in range(2, k+1):
        T_k.append(chebyshev_recurrence(T_k[-1], T_k[-2], X))

    return T_k


def sparse_to_tuple(sparse_mx):
    if not sp.isspmatrix_coo(sparse_mx):
        sparse_mx = sparse_mx.tocoo()
    coords = np.vSPACEL((sparse_mx.row, sparse_mx.col)).transpose()
    values = sparse_mx.data
    shape = sparse_mx.shape
    return coords, values, shape


from . import model
from . import data_utils
import numpy as np
import pandas as pd
import scanpy as sc
import scvi
import anndata
from . import metrics
import matplotlib.pyplot as plt
import os
from copy import deepcopy
import logging
import itertools
from functools import partial
from tqdm import tqdm
from time import strftime, localtime

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import TensorDataset, DataLoader, RandomSampler, BatchSampler
from torch.nn.parameter import Parameter
from torch.nn.modules.module import Module
from torch.utils.data.dataloader import default_collate

def guassian_kernel(source, target):
    n_samples = int(source.shape[0])+int(target.shape[0])
    total = np.concatenate((source, target), axis=0)
    total0 = np.expand_dims(total,axis=0)
    total0= np.broadcast_to(total0,[int(total.shape[0]), int(total.shape[0]), int(total.shape[1])])
    total1 = np.expand_dims(total,axis=1)
    total1=np.broadcast_to(total1,[int(total.shape[0]), int(total.shape[0]), int(total.shape[1])])
    L2_distance_square = np.cumsum(np.square(total0-total1),axis=2)
    bandwidth = np.sum(L2_distance_square) / (n_samples**2-n_samples)
    kernel_val = np.exp(-L2_distance_square / bandwidth)
    return kernel_val

def MMD(source, target):
    batch_size = int(source.shape[0])
    kernels = guassian_kernel(source, target)
    loss = 0
    for i in range(batch_size):
        s1, s2 = i, (i + 1) % batch_size
        t1, t2 = s1 + batch_size, s2 + batch_size
        loss += kernels[s1, s2] + kernels[t1, t2]
        loss -= kernels[s1, t2] + kernels[s2, t1]
    n_loss= loss / float(batch_size)
    return np.mean(n_loss)

def compute_kernel(x, y):
    x_size = x.size(0)
    y_size = y.size(0)
    dim = x.size(1)
    tiled_x = torch.tile(torch.reshape(x, (x_size, 1, dim)), (1, y_size, 1))
    tiled_y = torch.tile(torch.reshape(y, (1, y_size, dim)), (x_size, 1, 1))
    return torch.exp(-torch.square(tiled_x - tiled_y).mean(2) / dim)

def compute_mmd(x, y):
    x_kernel = compute_kernel(x, x)
    y_kernel = compute_kernel(y, y)
    xy_kernel = compute_kernel(x, y)
    return x_kernel.mean() + y_kernel.mean() - 2 * xy_kernel.mean()

class PredictionModel(nn.Module):
    def __init__(
        self, 
        input_dims,
        latent_dims,
        hidden_dims,
        celltype_dims,
        dropout=0.5,
    ):
        super(PredictionModel, self).__init__()
        
        self.encoder = nn.Sequential(
            nn.Linear(input_dims, hidden_dims),
            nn.LeakyReLU(),
            nn.BatchNorm1d(hidden_dims),
            nn.Linear(hidden_dims, hidden_dims),
            nn.LeakyReLU(),
            nn.BatchNorm1d(hidden_dims),
            nn.Linear(hidden_dims, hidden_dims),
            nn.LeakyReLU(),
            nn.BatchNorm1d(hidden_dims),
            nn.Dropout(dropout),
            nn.Linear(hidden_dims, latent_dims)
        )
        self.decoder = nn.Sequential(
            nn.Linear(latent_dims, hidden_dims),
            nn.LeakyReLU(),
            nn.BatchNorm1d(hidden_dims),
            nn.Linear(hidden_dims, hidden_dims),
            nn.LeakyReLU(),
            nn.BatchNorm1d(hidden_dims),
            nn.Linear(hidden_dims, hidden_dims),
            nn.LeakyReLU(),
            nn.BatchNorm1d(hidden_dims),
            nn.Dropout(dropout),
            nn.Linear(hidden_dims, input_dims)
        )
        self.pred = nn.Sequential(
            nn.Linear(latent_dims, hidden_dims),
            nn.LeakyReLU(),
            nn.Linear(hidden_dims, celltype_dims),
            nn.Softmax(dim=1),
        )
                
        nn.init.kaiming_normal_(self.encoder[0].weight)
        nn.init.kaiming_normal_(self.encoder[3].weight)
        nn.init.kaiming_normal_(self.encoder[6].weight)
        nn.init.xavier_uniform_(self.encoder[-1].weight)
        
        nn.init.kaiming_normal_(self.decoder[0].weight)
        nn.init.kaiming_normal_(self.decoder[3].weight)
        nn.init.kaiming_normal_(self.decoder[6].weight)
        nn.init.xavier_uniform_(self.decoder[-1].weight)

        nn.init.kaiming_normal_(self.pred[0].weight)
        nn.init.xavier_uniform_(self.pred[2].weight)
        
    @staticmethod
    def l2_activate(x,dim):
        
        def scale(z):
            zmax = z.max(1, keepdims=True).values
            zmin = z.min(1, keepdims=True).values
            z_std = torch.nan_to_num(torch.div(z - zmin,(zmax - zmin)),0)
            return z_std
        
        x = scale(x)
        x = F.normalize(x, p=2, dim=1)
        return x

    def forward(self, x):
        z = self.l2_activate(self.encoder(x),dim=1)
        pred = self.pred(z)
        decoded = self.decoder(z)
        return z, pred, decoded

class SpointModel():
    def __init__(
        self, 
        st_ad, 
        sm_ad, 
        clusters, 
        used_genes, 
        spot_names, 
        st_batch_key=None,
        scvi_layers=2,
        scvi_latent=64,
        scvi_gene_likelihood='zinb',
        scvi_dispersion='gene-batch',
        latent_dims=32, 
        hidden_dims=512,
        l1=0.01,
        l2=0.01,
        sm_lr=3e-4,
        st_lr=3e-4,
        use_gpu=None,
        seed=42
    ):
        if ((use_gpu is None) or (use_gpu is True)) and (torch.cuda.is_available()):
            self.device = 'cuda'
        else:
            self.device = 'cpu'
        self.use_gpu = use_gpu
        
        self.st_ad = st_ad
        self.sm_ad = sm_ad
        self.scvi_dims=64
        self.spot_names = spot_names
        self.used_genes = used_genes
        self.clusters = clusters
        self.st_batch_key = st_batch_key
        self.scvi_layers = scvi_layers
        self.scvi_latent = scvi_latent
        self.scvi_gene_likelihood = scvi_gene_likelihood
        self.scvi_dispersion = scvi_dispersion
        self.mse_loss_func = F.mse_loss
        self.kl_infer_loss_func = partial(self.kl_divergence, dim=1)
        self.kl_rec_loss_func = partial(self.kl_divergence, dim=0)
        self.cosine_infer_loss_func = partial(F.cosine_similarity, dim=1)
        self.cosine_rec_loss_func = partial(F.cosine_similarity, dim=0)
        self.l1 = l1
        self.l2 = l2
        self.model = PredictionModel(scvi_latent,latent_dims,hidden_dims,len(clusters)).to(self.device)
        self.sm_optimizer = optim.Adam(self.model.parameters(),lr=sm_lr)
        self.st_optimizer = optim.Adam(list(self.model.encoder.parameters())+list(self.model.decoder.parameters()),lr=st_lr)
        self.best_path = None
        self.history = pd.DataFrame(columns = ['sm_train_rec_loss','sm_train_infer_loss','sm_train_total_loss','sm_test_rec_loss','sm_test_infer_loss','sm_test_total_loss','st_train_rec_loss','st_test_rec_loss','is_best'])
        self.batch_size = None
        self.seed = seed
        
    @staticmethod
    def kl_divergence(y_true, y_pred, dim=0):
        y_pred = torch.clip(y_pred, torch.finfo(torch.float32).eps)
        y_true = y_true.to(y_pred.dtype)
        y_true = torch.nan_to_num(torch.div(y_true, y_true.sum(dim, keepdims=True)),0)
        y_pred = torch.nan_to_num(torch.div(y_pred, y_pred.sum(dim, keepdims=True)),0)
        y_true = torch.clip(y_true, torch.finfo(torch.float32).eps, 1)
        y_pred = torch.clip(y_pred, torch.finfo(torch.float32).eps, 1)
        return torch.mul(y_true, torch.log(torch.nan_to_num(torch.div(y_true, y_pred)))).mean(dim)
        
    def get_scvi_latent(
        self,
        n_layers,
        n_latent,
        gene_likelihood,
        dispersion,
        max_epochs,
        early_stopping,
        batch_size,
    ):
        if self.st_batch_key is not None:
            if 'simulated' in self.st_ad.obs[self.st_batch_key]:
                raise ValueError(f'obs[{self.st_batch_key}] cannot include "real".')
            self.st_ad.obs["batch"] = self.st_ad.obs[self.st_batch_key].astype(str)
            self.sm_ad.obs["batch"] = 'simulated'
        else:
            self.st_ad.obs["batch"] = 'real'
            self.sm_ad.obs["batch"] = 'simulated'

        adata = sc.concat([self.st_ad,self.sm_ad])
        adata.layers["counts"] = adata.X.copy()

        scvi.model.SCVI.setup_anndata(
            adata,
            layer="counts",
            batch_key="batch"
        )

        vae = scvi.model.SCVI(adata, n_layers=n_layers, n_latent=n_latent, gene_likelihood=gene_likelihood,dispersion=dispersion)
        vae.train(max_epochs=max_epochs,early_stopping=early_stopping,batch_size=batch_size,use_gpu=self.use_gpu)
        adata.obsm["X_scVI"] = vae.get_latent_representation()

        st_scvi_ad = anndata.AnnData(adata[adata.obs['batch'] != 'simulated'].obsm["X_scVI"])
        sm_scvi_ad = anndata.AnnData(adata[adata.obs['batch'] == 'simulated'].obsm["X_scVI"])

        st_scvi_ad.obs = self.st_ad.obs
        st_scvi_ad.obsm = self.st_ad.obsm

        sm_scvi_ad.obs = self.sm_ad.obs
        sm_scvi_ad.obsm = self.sm_ad.obsm
        
        sm_scvi_ad = data_utils.check_data_type(sm_scvi_ad)
        st_scvi_ad = data_utils.check_data_type(st_scvi_ad)

        self.sm_data = sm_scvi_ad.X
        self.sm_labels = sm_scvi_ad.obsm['label'].values
        self.st_data = st_scvi_ad.X
        
        return sm_scvi_ad,st_scvi_ad

    
    def build_dataset(self, batch_size, device):
        x_train,y_train,x_test,y_test = data_utils.split_shuffle_data(np.array(self.sm_data,dtype=np.float32),np.array(self.sm_labels,dtype=np.float32))
        
        x_train = torch.tensor(x_train).to(device)
        y_train = torch.tensor(y_train).to(device)
        x_test = torch.tensor(x_test).to(device)
        y_test = torch.tensor(y_test).to(device)
        st_data = torch.tensor(self.st_data).to(device)
        
        self.sm_train_ds = TensorDataset(x_train, y_train)
        self.sm_test_ds = TensorDataset(x_test,y_test)
        self.st_ds = TensorDataset(st_data)
        
        self.sm_train_batch_size = min(len(self.sm_train_ds), batch_size)
        self.sm_test_batch_size = min(len(self.sm_test_ds), batch_size)
        self.st_batch_size = min(len(self.st_ds), batch_size)
        
        g = torch.Generator()
        g.manual_seed(self.seed)
        self.sm_train_sampler = BatchSampler(RandomSampler(self.sm_train_ds, generator=g), batch_size=self.sm_train_batch_size, drop_last=True)
        self.sm_test_sampler = BatchSampler(RandomSampler(self.sm_test_ds, generator=g), batch_size=self.sm_test_batch_size, drop_last=True)
        self.st_sampler = BatchSampler(RandomSampler(self.st_ds, generator=g), batch_size=self.st_batch_size, drop_last=True)
    
    def train_sm(self, data, labels, rec_w=0.5, infer_w=1):
        self.model.train()
        self.sm_optimizer.zero_grad()
        latent, predictions, rec_data = self.model(data)
        rec_loss = self.kl_rec_loss_func(data, rec_data).mean() - self.cosine_rec_loss_func(data,rec_data).mean()
        infer_loss = self.kl_infer_loss_func(labels,predictions).mean() - self.cosine_infer_loss_func(labels,predictions).mean()

        regularization_params = torch.cat([
            torch.cat([x.view(-1) for x in self.model.encoder[0].parameters()]),
            torch.cat([x.view(-1) for x in self.model.encoder[3].parameters()]),
            torch.cat([x.view(-1) for x in self.model.encoder[6].parameters()]),
            torch.cat([x.view(-1) for x in self.model.decoder[0].parameters()]),
            torch.cat([x.view(-1) for x in self.model.decoder[3].parameters()]),
            torch.cat([x.view(-1) for x in self.model.decoder[6].parameters()]),
            torch.cat([x.view(-1) for x in self.model.pred[0].parameters()]),
        ])
        l1_regularization = self.l1 * torch.norm(regularization_params, 1)
        l2_regularization = self.l2 * torch.norm(regularization_params, 2)
        l1_l2_loss = l1_regularization + l2_regularization
        
        loss = rec_w*rec_loss + infer_w*infer_loss # + l1_l2_loss
        loss.backward()
        self.sm_optimizer.step()
        return latent , loss, rec_loss, infer_loss

    def train_st(self, data):
        self.model.train()
        self.st_optimizer.zero_grad()
        latent, predictions, rec_data = self.model(data)
        rec_loss = self.kl_rec_loss_func(data,rec_data).mean() - self.cosine_rec_loss_func(data,rec_data).mean()
        regularization_params = torch.cat([
            torch.cat([x.view(-1) for x in self.model.encoder[0].parameters()]),
            torch.cat([x.view(-1) for x in self.model.encoder[3].parameters()]),
            torch.cat([x.view(-1) for x in self.model.encoder[6].parameters()]),
            torch.cat([x.view(-1) for x in self.model.decoder[0].parameters()]),
            torch.cat([x.view(-1) for x in self.model.decoder[3].parameters()]),
            torch.cat([x.view(-1) for x in self.model.decoder[6].parameters()]),
            torch.cat([x.view(-1) for x in self.model.pred[0].parameters()]),
        ])
        l1_regularization = self.l1 * torch.norm(regularization_params, 1)
        l2_regularization = self.l2 * torch.norm(regularization_params, 2)
        l1_l2_loss = l1_regularization + l2_regularization
        loss = rec_loss # + l1_l2_loss
        loss.backward()
        self.st_optimizer.step()
        return latent , rec_loss
        
    def test_sm(self, data, labels, rec_w=0.5, infer_w=1):
        self.model.eval()
        latent, predictions, rec_data = self.model(data)
        rec_loss = rec_w*(self.kl_rec_loss_func(data,rec_data).mean() - self.cosine_rec_loss_func(data,rec_data).mean())
        infer_loss = infer_w*(self.kl_infer_loss_func(labels,predictions).mean() - self.cosine_infer_loss_func(labels,predictions).mean())
        loss = rec_loss + infer_loss
        return latent, loss, rec_loss, infer_loss
        
    def test_st(self, data):
        self.model.eval()
        latent, predictions, rec_data = self.model(data)
        rec_loss = self.kl_rec_loss_func(data,rec_data).mean() - self.cosine_rec_loss_func(data,rec_data).mean()
        return latent , rec_loss
    
    def train_model_by_step(
        self,
        max_steps=5000,
        save_mode='best',
        save_path=None,
        prefix=None,
        sm_step=4,
        st_step=1,
        test_step_gap=10,
        convergence=0.001,
        early_stop=True,
        early_stop_max=20,
        sm_lr=None,
        st_lr=None,
        rec_w=0.5, 
        infer_w=1,
    ):
        if len(self.history) > 0:
            best_ind = np.where(self.history['is_best'] == 'True')[0][-1]
            best_loss = self.history['sm_test_infer_loss'][best_ind]
            best_rec_loss = self.history['st_test_rec_loss'][best_ind]
        else:
            best_loss = np.inf
            best_rec_loss = np.inf
        early_stop_count = 0
        if sm_lr is not None:
            for g in self.sm_optimizer.param_groups:
                g['lr'] = sm_lr
        if st_lr is not None:
            for g in self.st_optimizer.param_groups:
                g['lr'] = st_lr

        pbar = tqdm(range(max_steps))
        mmd_loss = 0
        sm_trainr_iter = itertools.cycle(self.sm_train_sampler)
        st_iter = itertools.cycle(self.st_sampler)
        sm_shuffle_step = max(int(len(self.sm_train_ds)/(self.sm_train_batch_size*sm_step)),1)
        st_shuffle_step = max(int(len(self.st_ds)/(self.st_batch_size*st_step)),1)
        for step in pbar:
            if step % sm_shuffle_step == 0:
                sm_train_iter = itertools.cycle(self.sm_train_sampler)
            if step % st_shuffle_step == 0:
                st_iter = itertools.cycle(self.st_sampler)
            for i in range(st_step):
                exp = self.st_ds[next(st_iter)][0]
                st_latent, st_train_rec_loss = self.train_st(exp)
            for i in range(sm_step):
                exp, proportion = self.sm_train_ds[next(sm_train_iter)]
                sm_latent, sm_train_total_loss, sm_train_rec_loss, sm_train_infer_loss = self.train_sm(exp, proportion, rec_w=rec_w, infer_w=infer_w)
            # mmd_loss = compute_mmd(st_latent,sm_latent)*m_w

            if step % test_step_gap == 0:
                for ind in self.st_sampler:
                    exp = self.st_ds[ind][0]
                    st_latent, st_test_rec_loss = self.test_st(exp)
                for ind in self.sm_test_sampler:
                    exp, proportion = self.sm_test_ds[ind]
                    sm_latent, sm_test_total_loss, sm_test_rec_loss, sm_test_infer_loss = self.test_sm(exp,proportion, rec_w=rec_w, infer_w=infer_w)
                # mmd_loss = compute_mmd(st_latent,sm_latent)*m_w

                current_infer_loss = sm_test_infer_loss.item()

                best_flag='False'
                if best_loss - current_infer_loss > convergence:
                    if best_loss > current_infer_loss:
                        best_loss = current_infer_loss
                    best_flag='True'
                    # print('### Update best model')
                    early_stop_count = 0
                    old_best_path = self.best_path
                    if prefix is not None:
                        self.best_path = os.path.join(save_path,prefix+'_'+f'celleagle_weights_step{step}.h5')
                    else:
                        self.best_path = os.path.join(save_path,f'celleagle_weights_step{step}.h5')
                    if save_mode == 'best':
                        if old_best_path is not None:
                            if os.path.exists(old_best_path):
                                os.remove(old_best_path)
                        torch.save(self.model.state_dict(), self.best_path)
                else:
                    early_stop_count += 1
                    
                if save_mode == 'all':
                    if prefix != '':
                        self.best_path = os.path.join(save_path,prefix+'_'+f'celleagle_weights_step{step}.h5')
                    else:
                        self.best_path = os.path.join(save_path,f'celleagle_weights_step{step}.h5')
                    torch.save(self.model.state_dict(), self.best_path)
                    
                self.history = self.history.append({
                    'sm_train_rec_loss':sm_train_rec_loss.item(),
                    'sm_train_infer_loss':sm_train_infer_loss.item(),
                    'sm_train_total_loss':sm_train_total_loss.item(),
                    'sm_test_rec_loss':sm_test_rec_loss.item(),
                    'sm_test_infer_loss':sm_test_infer_loss.item(),
                    'sm_test_total_loss':sm_test_total_loss.item(),
                    'st_train_rec_loss':st_train_rec_loss.item(),
                    'st_test_rec_loss':st_test_rec_loss.item(),
                    'is_best':best_flag
                }, ignore_index=True)

                pbar.set_description(f"Step {step + 1}: test inf loss={sm_test_infer_loss.item():.3f}, train inf loss={sm_train_infer_loss.item():.3f}, test rec loss={sm_test_rec_loss.item():.3f}, train rec loss={sm_train_rec_loss.item():.3f}, st test rec loss={st_test_rec_loss.item():.3f}",refresh=True)
                
                if (early_stop_count > early_stop_max) and early_stop:
                    print('Stop trainning because of loss convergence')
                    break
            
    def train(
        self,
        max_steps=20000,
        save_mode='best',
        save_path=None,
        prefix=None,
        sm_step=4,
        st_step=1,
        test_step_gap=10,
        convergence=0.001,
        early_stop=True,
        early_stop_max=200,
        sm_lr=None,
        st_lr=None,
        batch_size=4096,
        rec_w=0.5, 
        infer_w=1,
        scvi_max_epochs=100,
        scvi_early_stopping=True,
        scvi_batch_size=4096,
    ):
        """Training Spoint model.
        
        Obtain latent feature from scVI then feed in Spoint model for training until loss convengence.

        Args:
            max_steps: The max step of training. The training process will be stop when achive max step.
            save_mode: A string determinates how the model is saved. It must be one of 'best' and 'all'.
            save_path: A string representing the path directory where the model is saved.
            prefix: A string added to the prefix of file name of saved model.
            convergence: The threshold of early stop.
            early_stop: If True, turn on early stop.
            early_stop_max: The max steps of loss difference less than convergence.
            sm_lr: Learning rate for simulated data.
            st_lr: Learning rate for spatial transcriptomic data.
            disc_lr: Learning rate of discriminator.
            batch_size: Batch size of the data be feeded in model once.
            rec_w: The weight of reconstruction loss.
            infer_w: The weig ht of inference loss.
            m_w: The weight of MMD loss.
            scvi_max_epochs: The max epoch of scVI.
            scvi_batch_size: The batch size of scVI.
        
        Returns:
            ``None``
        """

        self.get_scvi_latent(
            n_layers=self.scvi_layers,
            n_latent=self.scvi_latent,
            gene_likelihood=self.scvi_gene_likelihood,
            dispersion=self.scvi_dispersion,
            max_epochs=scvi_max_epochs,
            early_stopping=scvi_early_stopping,
            batch_size=scvi_batch_size,
        )
        self.build_dataset(batch_size, self.device)
        if save_path is None:
            save_path = 'Spoint_models_'+strftime("%Y%m%d%H%M%S",localtime())
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        self.train_model_by_step(
            max_steps=max_steps,
            save_mode=save_mode,
            save_path=save_path,
            prefix=prefix,
            sm_step=sm_step,
            st_step=st_step,
            test_step_gap=test_step_gap,
            convergence=convergence,
            early_stop=early_stop,
            early_stop_max=early_stop_max,
            sm_lr=sm_lr,
            st_lr=st_lr,
            rec_w=rec_w, 
            infer_w=infer_w,
        )
    
    def eval_model(self,model_path=None,use_best_model=True,batch_size=4096,metric='pcc'):
        if metric=='pcc':
            metric_name = 'PCC'
            func = metrics.pcc
        if metric=='spcc':
            metric_name = 'SPCC'
            func = metrics.spcc
        if metric=='mae':
            metric_name = 'MAE'
            func = metrics.mae
        if metric=='js':
            metric_name = 'JS'
            func = metrics.js
        if metric=='rmse':
            metric_name = 'RMSE'
            func = metrics.rmse
        if metric=='ssim':
            metric_name = 'SSIM'
            func = metrics.ssim
        
        if model_path is not None:
            self.model.load_state_dict(torch.load(model_path))
        elif use_best_model:
            self.model.load_state_dict(torch.load(self.best_path))
        model.eval()
        pre = []
        prop = []
        for exp_batch, prop_batch in self.sm_test_dataloader:
            latent_tmp, pre_tmp, _ = self.model(exp_batch)
            pre.extend(pre_tmp.cpu().detach().numpy())
            prop.extend(prop_batch.cpu().detach().numpy())
        pre = np.array(pre)
        prop = np.array(prop)
        metric_list = []
        for i,c in enumerate(self.clusters):
            metric_list.append(func(pre[:,i],prop[:,i]))
        print('### Evaluate model with simulation data')
        for i in range(len(metric_list)):
            print(f'{metric_name} of {self.clusters[i]}, {metric_list[i]}')
            
    def plot_training_history(self,save=None,return_fig=False,show=True,dpi=300):
        if len(self.history) > 0:
            fig, ax = plt.subplots()
            plt.plot(np.arange(len(self.history)), self.history['sm_test_infer_loss'], label='sm_test_infer_loss')
            plt.plot(np.arange(len(self.history)), self.history['st_test_rec_loss'], label='st_test_rec_loss')
            plt.xlabel('Epochs')
            plt.ylabel('Losses')
            plt.title('Training history')
            plt.legend()
            if save is not None:
                plt.savefig(save,bbox_inches='tight',dpi=dpi)
            if show:
                plt.show()
            plt.close()
            if return_fig:
                return fig
        else:
            print('History is empty, training model first')
            

    def deconv_spatial(self,st_data=None,min_prop=0.01,model_path=None,use_best_model=True,add_obs=True,add_uns=True):
        """Deconvolute spatial transcriptomic data.
        
        Using well-trained Spoint model to predict the cell type porportion of spots in spatial transcriptomic data.
        
        Args:
            st_data: An AnnData object of spatial transcriptomic data to be deconvolute.
            min_prop: A threshold value below which the predicted value will be set to 0. 
            model_path: A string representing the path of saved model file.
            use_best_model: If True, the model with the least loss will be used, otherwise, the last trained model will be used.
            add_obs: If True, the predicted results will be writen to the obs of input AnnData object of spatial transcriptomic data.
            add_uns: If True, the name of predicted cell types will be writen to the uns of input AnnData object of spatial transcriptomic data
        
        Returns:
            A ``DataFrame`` contained deconvoluted results. Each row representing a spot, and each column representing a cell type.
        """
        if st_data is None:
            st_data = self.st_data
        st_data = torch.tensor(st_data).to(self.device)
        if model_path is not None:
            self.model.load_state_dict(torch.load(model_path))
        elif use_best_model:
            self.model.load_state_dict(torch.load(self.best_path))
        self.model.to(self.device)
        self.model.eval()
        latent, pre, _ = self.model(st_data)
        pre = pre.cpu().detach().numpy()
        pre[pre < min_prop] = 0
        pre = pd.DataFrame(pre,columns=self.clusters,index=self.st_ad.obs_names)
        self.st_ad.obs[pre.columns] = pre.values
        self.st_ad.uns['celltypes'] = list(pre.columns)
        return pre


import scipy.sparse as sp
import numpy as np
from scipy.sparse.linalg import eigsh, ArpackNoConvergence
import torch

def normalize_adj(adj, symmetric=True):
    if symmetric:
        d = sp.diags(np.power(np.array(adj.sum(1)), -0.5).flatten(), 0)
        a_norm = adj.dot(d).transpose().dot(d).tocsr()
    else:
        d = sp.diags(np.power(np.array(adj.sum(1)), -1).flatten(), 0)
        a_norm = d.dot(adj).tocsr()
    return a_norm

def normalized_laplacian(adj, symmetric=True):
    adj_normalized = normalize_adj(adj, symmetric)
    laplacian = sp.eye(adj.shape[0]) - adj_normalized
    return laplacian


def rescale_laplacian(laplacian):
    try:
        print('Calculating largest eigenvalue of normalized graph Laplacian...')
        largest_eigval = eigsh(laplacian, 1, which='LM', return_eigenvectors=False)[0]
    except ArpackNoConvergence:
        print('Eigenvalue calculation did not converge! Using largest_eigval=2 instead.')
        largest_eigval = 2

    scaled_laplacian = (2. / largest_eigval) * laplacian - sp.eye(laplacian.shape[0])
    return scaled_laplacian


def chebyshev_polynomial(X, k):
    """Calculate Chebyshev polynomials up to order k. Return a list of sparse matrices."""
    print("Calculating Chebyshev polynomials up to order {}...".format(k))

    T_k = list()
    T_k.append(sp.eye(X.shape[0]).tocsr())
    T_k.append(X)

    def chebyshev_recurrence(T_k_minus_one, T_k_minus_two, X):
        X_ = sp.csr_matrix(X, copy=True)
        return 2 * X_.dot(T_k_minus_one) - T_k_minus_two

    for i in range(2, k+1):
        T_k.append(chebyshev_recurrence(T_k[-1], T_k[-2], X))

    return T_k

def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    return torch.sparse.FloatTensor(indices, values, shape)


from . import data_utils
from . import base_model
from . import data_downsample
from . import data_augmentation
from . import spatial_simulation
import numpy as np
import torch
from scipy.sparse import csr_matrix
import numba
import logging
import random
# logging.basicConfig(level=print,
#                     format='%(asctime)s %(levelname)s %(message)s',
#                     datefmt='%m-%d %H:%M')
# logging.getLogger().setLevel(print)

def init_model(
    sc_ad,
    st_ad,
    celltype_key,
    sc_genes=None,
    st_genes=None,
    used_genes=None,
    deg_method:str='wilcoxon',
    n_top_markers:int=200,
    n_top_hvg:int=None,
    log2fc_min=0.5, 
    pval_cutoff=0.01, 
    pct_diff=None, 
    pct_min=0.1,
    st_batch_key=None,
    sm_size:int=500000,
    cell_counts=None,
    clusters_mean=None,
    cells_mean=10,
    cells_min=1,
    cells_max=20,
    scvi_layers=2,
    scvi_latent=64,
    scvi_gene_likelihood='zinb',
    scvi_dispersion='gene-batch',
    latent_dims=32, 
    hidden_dims=512,
    n_threads=4,
    seed=42,
    use_gpu=None
):
    """Initialize Spoint model.
    
    Given specific data and parameters to initialize Spoint model.

    Args:
        sc_ad: An AnnData object representing single cell reference.
        st_ad: An AnnData object representing spatial transcriptomic data.
        celltype_key: A string representing cell types annotation columns in obs of single cell reference.
        sc_genes: A sequence of strings containing genes of single cell reference used in Spoint model. Only used when ``used_genes`` is None.
        st_genes: A sequence of strings containing genes of spatial transcriptomic data used in Spoint model. Only used when ``used_genes`` is None.
        used_genes: A sequence of strings containing genes used in Spoint model.
        deg_method: A string passed to method parameter of scanpy.tl.rank_genes_groups.
        n_top_markers: The number of differential expressed genes in each cell type of single cell reference used in Spoint model.
        n_top_hvg: The number of highly variable genes of spatial transcriptomic data used in Spoint model.
        log2fc_min: The threshold of log2 fold-change used for filtering differential expressed genes of single cell reference.
        pval_cutoff: The threshold of p-value used for filtering differential expressed genes of single cell reference.
        pct_min: The threshold of precentage of expressed cells used for filtering differential expressed genes of single cell reference.
        st_batch_key: A column name in obs of spatial transcriptomic data representing batch groups of spatial transcriptomic data.
        sm_size: The number of simulated spots.
        hiddem_dims: The number of nodes of hidden layers in Spoint model.
        n_threads: The number of cpu core used for parallel.
    
    Returns:
        A ``SpointModel`` object.
    """
    
    print('Setting global seed:', seed)
    random.seed(seed)
    np.random.seed(seed)

    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    spatial_simulation.numba_set_seed(seed)
    numba.set_num_threads(n_threads)

    sc_ad = data_utils.normalize_adata(sc_ad,target_sum=1e4)
    st_ad = data_utils.normalize_adata(st_ad,target_sum=1e4)
    sc_ad, st_ad = data_utils.filter_model_genes(
        sc_ad,
        st_ad,
        celltype_key=celltype_key,
        deg_method=deg_method,
        n_top_markers=n_top_markers,
        n_top_hvg=n_top_hvg,
        used_genes=used_genes,
        sc_genes=sc_genes,
        st_genes=st_genes,
        log2fc_min=log2fc_min, 
        pval_cutoff=pval_cutoff, 
        pct_diff=pct_diff, 
        pct_min=pct_min
    )
    sm_ad = data_utils.generate_sm_adata(sc_ad,num_sample=sm_size,celltype_key=celltype_key,n_threads=n_threads,cell_counts=cell_counts,clusters_mean=clusters_mean,cells_mean=cells_mean,cells_min=cells_min,cells_max=cells_max)
    data_utils.downsample_sm_spot_counts(sm_ad,st_ad,n_threads=n_threads)

    model = base_model.SpointModel(
        st_ad,
        sm_ad,
        clusters = np.array(sm_ad.obsm['label'].columns),
        spot_names = np.array(st_ad.obs_names),
        used_genes = np.array(st_ad.var_names),
        st_batch_key=st_batch_key,
        scvi_layers=scvi_layers,
        scvi_latent=scvi_latent,
        scvi_gene_likelihood=scvi_gene_likelihood,
        scvi_dispersion=scvi_dispersion,
        latent_dims=latent_dims, 
        hidden_dims=hidden_dims,
        use_gpu=use_gpu,
        seed=seed
    )
    return model


import open3d as o3d
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import pdist, cdist

def get_alpha(loc,subset=200,k=200):
    dis = cdist(loc,loc[np.random.permutation(loc.shape[0])[:subset]])
    return np.median(np.sort(dis,axis=0)[200])

def smooth_mesh(mesh,taubin_iter=None,subdivide_iter=3,show=False):
    if taubin_iter is not None:
        print('filter with Taubin with 10 iterations')
        mesh = mesh.filter_smooth_taubin(number_of_iterations=taubin_iter)
        mesh.compute_vertex_normals()
    if subdivide_iter is not None:
        mesh = mesh.subdivide_loop(number_of_iterations=subdivide_iter)
        print(
            f'After subdivision it has {len(mesh.vertices)} vertices and {len(mesh.triangles)} triangles'
        )
    if show:
        o3d.visualization.draw_geometries([mesh], mesh_show_back_face=True)
    return mesh

def create_mesh(loc,alpha=None,show=False):
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(loc)
    if alpha is None:
        alpha = get_alpha(loc)
    print(f"alpha={alpha:.3f}")
    mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_alpha_shape(pcd, alpha)
    mesh.compute_vertex_normals()
    if show:
        o3d.visualization.draw_geometries([pcd,mesh], mesh_show_back_face=True)
    return mesh

def sample_in_mesh(mesh,xyz,num_sample=500000,save_sampled=None,save_surface=None):
    mesh_scene = o3d.t.geometry.TriangleMesh.from_legacy(mesh)

    # Create a scene and add the triangle mesh
    scene = o3d.t.geometry.RaycastingScene()
    _ = scene.add_triangles(mesh_scene)

    dx,dy,dz = xyz.max(0)-xyz.min(0)
    dd = np.math.pow(num_sample/((dx/dz)*(dy/dz)*(dz/dz)), 1/3)

    nx,ny,nz=int((dx/dz)*dd),int((dy/dz)*dd),int(dd)
    xmin,ymin,zmin=xyz.min(0)
    xmax,ymax,zmax=xyz.max(0)
    x_sampled = np.linspace(xmin-1e-7, xmax+1e-7, nx)
    y_sampled = np.linspace(ymin-1e-7, ymax+1e-7, ny)
    z_sampled = np.linspace(zmin-1e-7, zmax+1e-7, nz)

    xyz_sampled = []
    for x_s in x_sampled:
        for y_s in y_sampled:
            for z_s in z_sampled:
                xyz_sampled.append([x_s,y_s,z_s])

    xyz_sampled = np.array(xyz_sampled)

    query_point = o3d.core.Tensor(xyz_sampled, dtype=o3d.core.Dtype.Float32)

    # Compute distance of the query point from the surface
    unsigned_distance = scene.compute_distance(query_point)
    signed_distance = scene.compute_signed_distance(query_point)
    occupancy = scene.compute_occupancy(query_point)

    occupancy = np.array(occupancy)

    xyz_sampled_inmesh = np.array(xyz_sampled)[occupancy == 1]
    xyz_surface = np.asarray(mesh.vertices)
    if save_sampled is not None:
        np.save(save_sampled,xyz_sampled_inmesh)
    if save_surface is not None:
        np.save(save_surface,xyz_surface)
    return xyz_sampled_inmesh, xyz_surface

def convert_colors(val, cmap='Spectral_r'):
    val = (val - val.min())/(val.max()-val.min())
    cmap = plt.get_cmap(cmap)
    # point_colors = [cmap(val) for val in ano4_gpr]
    point_colors = [cmap(v) for v in val]
    point_colors = np.array(point_colors)[:,:3]
    return point_colors

def get_surface_colors(mesh, surface_expr, cmap='Spectral_r'):
    surface_expr = (surface_expr - surface_expr.min())/(surface_expr.max()-surface_expr.min())
    cmap = plt.get_cmap(cmap)
    point_colors = [cmap(val) for val in surface_expr]
    point_colors = np.array(point_colors)[:,:3]
    mesh.vertex_colors = o3d.utility.Vector3dVector(point_colors)
    
def save_view_parameters(pcd, filename, point_size=3):
    vis = o3d.visualization.Visualizer()
    vis.create_window()
    opt = vis.get_render_option()
    opt.mesh_show_back_face = True
    opt.point_size=point_size
    vis.add_geometry(pcd)
    vis.run()
    param = vis.get_view_control().convert_to_pinhole_camera_parameters()
    o3d.io.write_pinhole_camera_parameters(filename,param)
    vis.destroy_window()
    
def load_view_parameters(pcd, filename, point_size=3):
    vis = o3d.visualization.Visualizer()
    vis.create_window()
    ctr = vis.get_view_control()
    opt = vis.get_render_option()
    opt.mesh_show_back_face = True
    opt.point_size=point_size
    param = o3d.io.read_pinhole_camera_parameters(filename)
    vis.add_geometry(pcd)
    ctr.convert_from_pinhole_camera_parameters(param)
    vis.run()
    vis.destroy_window()


import matplotlib.pyplot as plt
from matplotlib.patches import Patch
from matplotlib.lines import Line2D
import matplotlib
import seaborn as sns
import pandas as pd
import numpy as np

def plot_3d(
    loc,
    val=None, 
    color=None, 
    figsize=(8,8), 
    return_fig=False, 
    elev=None, 
    azim=None, 
    xlim=None, 
    ylim=None, 
    zlim=None, 
    frameon=True, 
    save_path=None, 
    save_dpi=150, 
    show=True,
    *args,
    **kwargs
):
    """Plot all slices stacked in 3D
    
    Plot all slices stacked with a given number of subplots rows and columns. Spots/cells colored by spatial domain.

    Args:
        loc: An array of the coordinates of each spots/cells in all slices, which the first three columns are X-axis, Y-axis, Z-axis coordinates.
        val: The colors of each spots/cells given in loc use to plot. 
        color: The colors of each spots/cells given in loc use to plot. 
        figsize: Size of the figure.
        return_fig: Whether to return the figure. 
        elev: The elevation angle in the vertical plane in degrees. If ``None`` then the initial value as specified in the ``Axes3D`` constructor is used.
        azim: The azimuth angle in the horizontal plane in degrees. If ``None`` then the initial value as specified in the ``Axes3D`` constructor is used. 
        xlim: A tuple given the left and right xlims in X-axis. 
        ylim: A tuple given the left and right xlims in Y-axis. 
        zlim: A tuple given the left and right xlims in Z-axis. 
        frameon: Whether to hide the coordinate axes.  
        save_path: A string representing the path directory where the figure saved. 
        save_dpi: The resolution in dots per inch.
        show: Whether to show the figure. 
    
    Returns:
        A ``matplotlib.figure.Figure`` object
    """
    if 'marker' not in kwargs.keys():
        kwargs['marker'] = 'o'
    if 's' not in kwargs.keys():
        kwargs['s'] = 5
    if 'cmap' not in kwargs.keys():
        kwargs['cmap'] = 'Spectral_r'
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(projection='3d')
    if (elev is not None) and (azim is not None):
        ax.view_init(elev, azim)  # 设定视角
    if color is None:
        ax.scatter(loc[:,0], loc[:,1], loc[:,2],c=val,*args,**kwargs)
    else:
        ax.scatter(loc[:,0], loc[:,1], loc[:,2],c=color,*args,**kwargs)
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)
    if zlim is not None:
        ax.set_zlim(zlim)
    if not frameon:
        plt.axis('off')
    if save_path is not None:
        print(save_path)
        plt.savefig(save_path,dpi=save_dpi,bbox_inches='tight')
    if show:
        plt.show()
    else:
        plt.close()
    if return_fig:
        return fig

def plot_single_slice(adata, spatial_key, cluster_key, frameon=False, i=1, j=1, n=1, s=1):
    ind = np.sort(adata.obs[cluster_key].unique().copy())
    color = adata.uns[cluster_key+'_colors'].copy()
    dic = dict(zip(ind, color))
    col = adata.obs[cluster_key].replace(dic)
    ax = plt.subplot(i,j,n)
    if type(adata.obsm[spatial_key]) == pd.core.frame.DataFrame:
        plt.scatter(adata.obsm[spatial_key].iloc[:,0], adata.obsm[spatial_key].iloc[:,1], c=col, s=s, rasterized=True)
    if type(adata.obsm[spatial_key]) == np.ndarray:
        plt.scatter(adata.obsm[spatial_key][:,0], adata.obsm[spatial_key][:,1], c=col, s=s, rasterized=True)
    plt.axis('equal')
    if not frameon:
        plt.axis('off')

def plot_stacked_slices(
    ad_list, 
    spatial_key, 
    cluster_key, 
    legend=True, 
    frameon=False, 
    colors=None, 
    i=1, 
    j=1, 
    s=1
):
    """Plot all slices stacked
    
    Plot all slices stacked in one figure with a given number of subplots rows and columns. Spots/cells colored by spatial domain.

    Args:
        ad_list: A list of ``AnnData`` objects containing all slices.
        spatial_key: A string representing one key of ``obsm`` in AnnData object of all slices, containing the coordinates used to plot.
        cluster_key: A string representing one column of ``obs`` in AnnData object of all slices, containing the spatial domain information used to plot.
        legend: Whether to display the legend.
        frameon: Whether to hide the coordinate axes. 
        colors: A list of colors for each spatial domain to plot. If ``None``, it will default to ``tab10`` or ``tab20`` accoording to the number of spatial domains.
        i: Number of rows of the subplots.
        j: Number of columns of the subplots. If i=j=1, it will be a single figure.
        s: Size of points.
    
    Returns:
        ``None``
    """
    
    clusters = []
    colored_num = 0
    for ad in ad_list:
        if f'{cluster_key}_colors' in ad.uns.keys():
            colored_num += 1
        clusters.extend(ad.obs[cluster_key].cat.categories)
    clusters = np.unique(clusters)
    if colored_num < len(ad_list):
        if colors is None:
            if len(clusters) > 10:
                colors = [matplotlib.colors.to_hex(c) for c in sns.color_palette('tab20',n_colors=len(clusters))]
            else:
                colors = [matplotlib.colors.to_hex(c) for c in sns.color_palette('tab10',n_colors=len(clusters))]
        color_map = pd.DataFrame(colors,index=clusters,columns=['color'])
        for ad in ad_list:
            ad.uns[f'{cluster_key}_color_map'] = color_map
            ad.uns[f'{cluster_key}_colors'] = [color_map.loc[c,'color'] for c in ad.obs[cluster_key].cat.categories]
    else:
        color_map = pd.DataFrame(index=clusters,columns=['color'])
        for ad in ad_list:
            color_map.loc[ad.obs[cluster_key].cat.categories,'color'] = ad.uns[f'{cluster_key}_colors']
            ad.uns[f'{cluster_key}_color_map'] = color_map
    if legend:
        legend_elements = [ Line2D([], [], marker='.', markersize=10, color=color_map.loc[i,'color'], linestyle='None', label=i) for i in color_map.index]
        for k in range(len(ad_list)):
            plot_single_slice(ad_list[k], spatial_key, cluster_key, frameon, i, j, n=np.min([k+1,i*j]), s=s)
        plt.legend(handles=legend_elements, loc='right')
        


from __future__ import print_function

from tensorflow.keras import activations, initializers, constraints
from tensorflow.keras import regularizers
from tensorflow.keras.layers import Layer
import tensorflow.keras.backend as K
import tensorflow as tf


class GraphConvolution(Layer):
    """Basic graph convolution layer as in https://arxiv.org/abs/1609.02907"""
    def __init__(self, units, support=1,
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(GraphConvolution, self).__init__(**kwargs)
        self.units = units
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        self.supports_masking = True

        self.support = support
        assert support >= 1

    def compute_output_shape(self, input_shapes):
        features_shape = input_shapes[0]
        output_shape = (features_shape[0], self.units)
        return output_shape  # (batch_size, output_dim)

    def build(self, input_shapes):
        features_shape = input_shapes[0]
        assert len(features_shape) == 2
        input_dim = features_shape[1]

        self.kernel = self.add_weight(shape=(input_dim * self.support,
                                             self.units),
                                      initializer=self.kernel_initializer,
                                      name='kernel',
                                      regularizer=self.kernel_regularizer,
                                      constraint=self.kernel_constraint)
        if self.use_bias:
            self.bias = self.add_weight(shape=(self.units,),
                                        initializer=self.bias_initializer,
                                        name='bias',
                                        regularizer=self.bias_regularizer,
                                        constraint=self.bias_constraint)
        else:
            self.bias = None
        self.built = True

    def call(self, inputs, mask=None):
        features = inputs[0]
        basis = inputs[1:]

        supports = list()
        for i in range(self.support):
#             supports.append(K.dot(basis[i], features))
#             supports.append(K.dot(tf.sparse.to_dense(basis[i]), features))
#             supports.append(K.dot(tf.sparse.to_dense(basis[i]), tf.sparse.to_dense(features)))
            supports.append(tf.sparse.sparse_dense_matmul(basis[i], features))
        supports = K.concatenate(supports, axis=1)
        output = K.dot(supports, self.kernel)

        if self.use_bias:
            output += self.bias
        return self.activation(output)

    def get_config(self):
        config = {'units': self.units,
                  'support': self.support,
                  'activation': activations.serialize(self.activation),
                  'use_bias': self.use_bias,
                  'kernel_initializer': initializers.serialize(
                      self.kernel_initializer),
                  'bias_initializer': initializers.serialize(
                      self.bias_initializer),
                  'kernel_regularizer': regularizers.serialize(
                      self.kernel_regularizer),
                  'bias_regularizer': regularizers.serialize(
                      self.bias_regularizer),
                  'activity_regularizer': regularizers.serialize(
                      self.activity_regularizer),
                  'kernel_constraint': constraints.serialize(
                      self.kernel_constraint),
                  'bias_constraint': constraints.serialize(self.bias_constraint)
        }

        base_config = super(GraphConvolution, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import numpy as np
from scipy.sparse import coo_matrix
import torch
import torch.nn.functional as F
from .pygcn_utils import *

def get_graph_inputs(celltype_ad_list):
    print('Generating GNN inputs...')
    A_list = []
    X_list = []
    for celltype_ad in celltype_ad_list:
        X_tmp = np.matrix(celltype_ad.X,dtype='float32')
        X_list.append(X_tmp)
        A_list.append(coo_matrix(celltype_ad.obsp['spatial_distances'],dtype='float32'))

    X_raw = np.concatenate(X_list)
    element_num = 0
    for A_tmp in A_list:
        element_num += A_tmp.shape[0]
    A = np.zeros((element_num,element_num),dtype='float32')
    loc_index = 0
    class_index = 0
    slice_class = []
    for A_tmp in A_list:
        A[loc_index:loc_index+A_tmp.shape[0],loc_index:loc_index+A_tmp.shape[0]] = A_tmp.toarray()
        slice_class = slice_class + [class_index]*A_tmp.shape[0]
        loc_index += A_tmp.shape[0]
        class_index += 1
    A = coo_matrix(A,dtype='float32')
    nb_mask = np.array(np.where((A>0).todense()))
    slice_class_onehot = F.one_hot(torch.tensor(slice_class)).float()
    return X_raw,A,nb_mask,slice_class_onehot

def get_graph_kernel(features,adj,k=2):
    features_scaled = (features-features.mean(0))/features.std(0)
    features_scaled = torch.tensor(features_scaled)
    SYM_NORM = False  # symmetric (True) vs. left-only (False) normalization
    L = normalized_laplacian(adj, SYM_NORM)
    L_scaled = rescale_laplacian(L)
    T_k = chebyshev_polynomial(L_scaled, k)
    support = k + 1
    graph = [features_scaled]+T_k
    for _i in range(len(graph))[1:]:
        graph[_i] = sparse_mx_to_torch_sparse_tensor(graph[_i])
    return features_scaled, graph, support

def split_train_test_idx(X,train_prop):
    rand_idx = np.random.permutation(X.shape[0])
    train_idx = rand_idx[:int(len(rand_idx)*train_prop)]
    test_idx = rand_idx[int(len(rand_idx)*train_prop):]
    return train_idx, test_idx


from scipy.stats import pearsonr, entropy, spearmanr
from scipy.spatial.distance import jensenshannon
from sklearn.metrics import mean_squared_error
import numpy as np

def pcc(x1,x2):
    return pearsonr(x1,x2)[0]

def spcc(x1,x2):
    return spearmanr(x1,x2)[0]

def rmse(x1,x2):
    return mean_squared_error(x1,x2,squared=False)

def mae(x1,x2):
    return np.mean(np.abs(x1-x2))

def js(x1,x2):
    return jensenshannon(x1,x2)

def kl(x1,x2):
    entropy(x1, x2)

def ssim(im1,im2,M=1):
    im1, im2 = im1/im1.max(), im2/im2.max()
    mu1 = im1.mean()
    mu2 = im2.mean()
    sigma1 = np.sqrt(((im1 - mu1) ** 2).mean())
    sigma2 = np.sqrt(((im2 - mu2) ** 2).mean())
    sigma12 = ((im1 - mu1) * (im2 - mu2)).mean()
    k1, k2, L = 0.01, 0.03, M
    C1 = (k1*L) ** 2
    C2 = (k2*L) ** 2
    C3 = C2/2
    l12 = (2*mu1*mu2 + C1)/(mu1 ** 2 + mu2 ** 2 + C1)
    c12 = (2*sigma1*sigma2 + C2)/(sigma1 ** 2 + sigma2 ** 2 + C2)
    s12 = (sigma12 + C3)/(sigma1*sigma2 + C3)
    ssim = l12 * c12 * s12
    return ssim


import numba
import numpy as np

@numba.njit
def random_dropout(cell_expr,max_rate):
    non_zero_mask = np.where(cell_expr!=0)[0]
    zero_mask = np.random.choice(non_zero_mask,int(len(non_zero_mask)*np.float32(np.random.uniform(0,max_rate))))
    cell_expr[zero_mask] = 0
    return cell_expr

@numba.njit
def random_scale(cell_expr,max_val):
    scale_factor = np.float32(1+np.random.uniform(-max_val,max_val))
    cell_expr = cell_expr*scale_factor
    return cell_expr

@numba.njit
def random_shift(cell_expr,kth):
    shift_value = np.random.choice(np.array([1,0,-1]),1)[0]*np.unique(cell_expr)[int(np.random.uniform(0,kth)*len(np.unique(cell_expr)))]
    cell_expr[cell_expr != 0] = cell_expr[cell_expr != 0]+shift_value
    cell_expr[cell_expr < 0] = 0
    return cell_expr

@numba.njit(parallel=True)
def random_augment(mtx,max_rate=0.8,max_val=0.8,kth=0.2):
    for i in numba.prange(mtx.shape[0]):
        random_dropout(mtx[i,:],max_rate=max_rate)
        random_scale(mtx[i,:],max_val=max_val)
        random_shift(mtx[i,:],kth=kth)
    return mtx

@numba.njit
def random_augmentation_cell(cell_expr,max_rate=0.8,max_val=0.8,kth=0.2):
    cell_expr = random_dropout(cell_expr,max_rate=max_rate)
    cell_expr = random_scale(cell_expr,max_val=max_val)
    cell_expr = random_shift(cell_expr,kth=kth)
    return cell_expr


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

d = {}
with open("SPACEL/_version.py") as f:
    exec(f.read(), d)

setup(
    name="SPACEL",
    version=d["__version__"],
    author="Hao Xu",
    author_email="xuhaoustc@mail.ustc.edu.cn",
    description="SPACEL: characterizing spatial transcriptome architectures by deep-learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/QuKunLab/SPACEL",
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "pip",
        "rpy2==3.3.*",
        "squidpy",
        "scvi-tools",
        "scipy",
        "scikit-learn",
        "matplotlib",
        "seaborn",
        "scanpy",
        "numba",
        "gpytorch",
        "torchmetrics",
    ],
)


import subprocess
import os
import numpy as np

def auto_cuda_device():
    out = subprocess.getoutput('nvidia-smi -q -d Memory |grep -A4 GPU|grep Free')
    if len(out.split('\n')) > 1:
        memory_available = [int(x.split()[2]) for x in out.split('\n')]
        max_idx = np.where(memory_available == np.max(memory_available))[0]
        os.environ['CUDA_VISIBLE_DEVICES']=str(np.random.permutation(max_idx)[0])
        print('Using GPU:',os.environ['CUDA_VISIBLE_DEVICES'])
    else:
        raise ValueError('Invalid output from nvidia-smi.')

def set_environ_seed(seed=42):
    os.environ['TF_DETERMINISTIC_OPS'] = '1'
    os.environ['PYTHONHASHSEED']=str(seed)
    print('Setting environment seed:',seed)
    
